"""BIP-39 mnemonic generator, encrypt/decrypt, ETH account derivation and keystore saving."""

import base64
import json
import os
from pathlib import Path
from typing import Dict

from bip_utils import (
    Bip39MnemonicGenerator,
    Bip39SeedGenerator,
    Bip39WordsNum,
    Bip44,
    Bip44Changes,
    Bip44Coins,
)
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from eth_account import Account
from pydantic import BaseModel

from iwa.core.constants import WALLET_PATH
from iwa.core.models import EncryptedData, EthereumAddress, StoredAccount

MNEMONIC_WORD_NUMBER = Bip39WordsNum.WORDS_NUM_24
SCRYPT_N = 2**14
SCRYPT_R = 8
SCRYPT_P = 1
SCRYPT_LEN = 32
AES_NONCE_LEN = 12
SALT_LEN = 16


class EncryptedMnemonic(EncryptedData):
    """EncryptedMnemonic"""

    def derive_key(self, password: bytes) -> bytes:
        """Derive a key from a password and salt using scrypt."""
        kdf = Scrypt(
            salt=base64.b64decode(self.kdf_salt),
            length=self.kdf_len,
            n=self.kdf_n,
            r=self.kdf_r,
            p=self.kdf_p,
        )
        return kdf.derive(password)

    def decrypt(self, password: str) -> str:
        """Decrypt an object."""
        # validate expected algorithms
        if self.kdf != "scrypt":
            raise ValueError(f"Unsupported kdf: {self.kdf}")
        if self.cipher != "aesgcm":
            raise ValueError("Unsupported cipher, expected 'aesgcm'")

        nonce = base64.b64decode(self.nonce)
        ct = base64.b64decode(self.ciphertext)

        # derive key using the parameters from the file
        key = self.derive_key(password.encode("utf-8"))
        aesgcm = AESGCM(key)
        pt = aesgcm.decrypt(nonce, ct, None)
        return pt.decode("utf-8")

    @classmethod
    def encrypt(cls, mnemonic: str, password: str) -> dict:
        """Encrypt a mnemonic with AES-GCM using a scrypt-derived key."""
        password_b = password.encode("utf-8")
        salt = os.urandom(SALT_LEN)
        # Create a temporary instance to use derive_key or use the class method if refactored
        # But derive_key is an instance method using instance attributes.
        # We should probably just use Scrypt directly here or refactor derive_key.
        # Since derive_key uses self.kdf_n etc, we can use the constants.

        kdf = Scrypt(
            salt=salt,
            length=SCRYPT_LEN,
            n=SCRYPT_N,
            r=SCRYPT_R,
            p=SCRYPT_P,
        )
        key = kdf.derive(password_b)

        aesgcm = AESGCM(key)
        nonce = os.urandom(AES_NONCE_LEN)
        ct = aesgcm.encrypt(nonce, mnemonic.encode("utf-8"), None)
        return {
            "kdf": "scrypt",
            "kdf_salt": base64.b64encode(salt).decode(),
            "kdf_n": SCRYPT_N,
            "kdf_r": SCRYPT_R,
            "kdf_p": SCRYPT_P,
            "kdf_len": SCRYPT_LEN,
            "cipher": "aesgcm",
            "nonce": base64.b64encode(nonce).decode(),
            "ciphertext": base64.b64encode(ct).decode(),
        }


class MnemonicStorage(BaseModel):
    """MnemonicStorage"""

    encrypted_mnemonic: EncryptedMnemonic
    accounts: Dict[EthereumAddress, StoredAccount] = {}

    @staticmethod
    def load(file_path: Path = WALLET_PATH) -> "MnemonicStorage":
        """Load"""
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)
        data["encrypted_mnemonic"] = EncryptedMnemonic(**data["encrypted_mnemonic"])
        data["accounts"] = {k: StoredAccount(**v) for k, v in data.get("accounts", {}).items()}
        return MnemonicStorage(**data)

    def save(self, file_path: Path = WALLET_PATH) -> None:
        """Save mnemonic storage to file with secure permissions."""
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(self.model_dump(), f, indent=4)
        # SECURITY: Restrict file permissions to owner only
        os.chmod(file_path, 0o600)


class MnemonicManager:
    """Manager for BIP-39 mnemonics and keystore operations.

    Provides methods to generate mnemonics, encrypt/decrypt them using
    scrypt + AES-GCM, derive Ethereum accounts (BIP-44), and save
    keystores to disk.

    Attributes:
        mnemonic_file (str): Default file path for the encrypted mnemonic.
        mnemonic_word_number (Bip39WordsNum): Number of words in the mnemonic.
        scrypt_*: Parameters for the scrypt KDF.
        aes_nonce_len (int): Nonce length for AES-GCM.
        salt_len (int): Salt length for scrypt.

    """

    def __init__(
        self,
        mnemonic_file: str = WALLET_PATH,
        mnemonic_word_number: Bip39WordsNum = MNEMONIC_WORD_NUMBER,
        scrypt_n: int = SCRYPT_N,
        scrypt_r: int = SCRYPT_R,
        scrypt_p: int = SCRYPT_P,
        scrypt_len: int = SCRYPT_LEN,
        aes_nonce_len: int = AES_NONCE_LEN,
        salt_len: int = SALT_LEN,
    ):
        """Initialize MnemonicManager with configuration parameters."""
        self.mnemonic_file = mnemonic_file
        self.mnemonic_word_number = mnemonic_word_number
        self.scrypt_n = scrypt_n
        self.scrypt_r = scrypt_r
        self.scrypt_p = scrypt_p
        self.scrypt_len = scrypt_len
        self.aes_nonce_len = aes_nonce_len
        self.salt_len = salt_len

    def derive_key(
        self,
        password: bytes,
        salt: bytes,
        n: int | None = None,
        r: int | None = None,
        p: int | None = None,
        length: int | None = None,
    ) -> bytes:
        """Derive a key from a password and salt using scrypt.

        Args:
            password (bytes): The password in bytes.
            salt (bytes): A random salt.
            n (int | None): CPU/memory cost factor.
            r (int | None): Block size parameter.
            p (int | None): Parallelization parameter.
            length (int | None): Desired key length in bytes.

        Returns:
            bytes: The derived key of length `self.scrypt_len`.

        """
        # use provided parameters or fall back to instance defaults
        n = n if n is not None else self.scrypt_n
        r = r if r is not None else self.scrypt_r
        p = p if p is not None else self.scrypt_p
        length = length if length is not None else self.scrypt_len

        kdf = Scrypt(
            salt=salt,
            length=length,
            n=n,
            r=r,
            p=p,
        )
        return kdf.derive(password)

    def encrypt_mnemonic(self, mnemonic: str, password: str) -> dict:
        """Encrypt a mnemonic with AES-GCM using a scrypt-derived key.

        Args:
            mnemonic (str): The mnemonic as plain text.
            password (str): Password used to derive the encryption key.

        Returns:
            dict: JSON-serializable object containing KDF params, nonce,
                  and ciphertext (all base64-encoded).

        """
        password_b = password.encode("utf-8")
        salt = os.urandom(self.salt_len)
        key = self.derive_key(password_b, salt)
        aesgcm = AESGCM(key)
        nonce = os.urandom(self.aes_nonce_len)
        ct = aesgcm.encrypt(nonce, mnemonic.encode("utf-8"), None)
        return {
            "kdf": "scrypt",
            "kdf_salt": base64.b64encode(salt).decode(),
            "kdf_n": self.scrypt_n,
            "kdf_r": self.scrypt_r,
            "kdf_p": self.scrypt_p,
            "kdf_len": self.scrypt_len,
            "cipher": "aesgcm",
            "nonce": base64.b64encode(nonce).decode(),
            "ciphertext": base64.b64encode(ct).decode(),
        }

    def decrypt_mnemonic(self, encobj: dict, password: str) -> str:
        """Decrypt an object previously created by `encrypt_mnemonic`.

        Args:
            encobj (dict): Object with KDF params, nonce and ciphertext
                           encoded in base64.
            password (str): Password to derive the decryption key.

        Returns:
            str: The mnemonic in plain text.

        """
        # validate expected algorithms
        kdf_name = encobj.get("kdf", "scrypt")
        if kdf_name != "scrypt":
            raise ValueError(f"Unsupported kdf: {kdf_name}")
        if encobj.get("cipher", "aesgcm") != "aesgcm":
            raise ValueError("Unsupported cipher, expected 'aesgcm'")

        salt = base64.b64decode(encobj["kdf_salt"])
        nonce = base64.b64decode(encobj["nonce"])
        ct = base64.b64decode(encobj["ciphertext"])

        # read kdf params from the encoded object, falling back to defaults
        n = int(encobj.get("kdf_n", self.scrypt_n))
        r = int(encobj.get("kdf_r", self.scrypt_r))
        p = int(encobj.get("kdf_p", self.scrypt_p))
        length = int(encobj.get("kdf_len", self.scrypt_len))

        # derive key using the parameters from the file
        key = self.derive_key(password.encode("utf-8"), salt, n=n, r=r, p=p, length=length)
        aesgcm = AESGCM(key)
        pt = aesgcm.decrypt(nonce, ct, None)
        return pt.decode("utf-8")

    def generate_and_store_mnemonic(
        self,
        password: str,
        out_file: str = None,
    ) -> str:
        """Generate a BIP-39 mnemonic, encrypt it and save to disk.

        Args:
            password (str): Password to encrypt the mnemonic.
            out_file (str): Destination file. Optional; if None this method
                            uses `self.mnemonic_file`.

        Returns:
            str: The plaintext mnemonic (returned so data is available).

        """
        out_file = out_file or self.mnemonic_file
        mnemonic = Bip39MnemonicGenerator().FromWordsNumber(self.mnemonic_word_number)
        mnemonic_str = mnemonic.ToStr()
        enc = self.encrypt_mnemonic(mnemonic_str, password)
        with open(out_file, "w", encoding="utf-8") as f:
            json.dump(enc, f, indent=2)
        os.chmod(out_file, 0o600)
        return mnemonic_str

    def load_and_decrypt_mnemonic(
        self,
        password: str,
        in_file: str = None,
    ) -> str:
        """Load and decrypt a mnemonic from a file.

        Args:
            password (str): Password to decrypt the mnemonic.
            in_file (str): File path with the encrypted object. Optional;
                           if None `self.mnemonic_file` is used.

        Returns:
            str: The plaintext mnemonic.

        """
        in_file = in_file or self.mnemonic_file
        with open(in_file, "r", encoding="utf-8") as f:
            enc = json.load(f)
        try:
            mnemonic = self.decrypt_mnemonic(enc, password)
            return mnemonic
        except InvalidTag as e:
            raise ValueError("Incorrect password") from e

    def derive_eth_accounts_from_mnemonic(
        self,
        password: str,
        n_accounts: int = 5,
    ):
        """Derive Ethereum accounts (BIP-44) from a BIP-39 mnemonic.

        Args:
            password (str): Password to decrypt the mnemonic.
            n_accounts (int): Number of accounts to derive.

        Returns:
            list: Dicts with keys 'index', 'address' and 'private_key_hex'.

        """
        mnemonic = self.load_and_decrypt_mnemonic(password)

        if mnemonic is None:
            return None

        accounts = []
        for i in range(n_accounts):
            # obtain private key (hex) for this index using helper
            priv_hex = self.derive_private_key_hex_from_mnemonic(mnemonic, i)
            priv_bytes = bytes.fromhex(priv_hex)
            acct = Account.from_key(priv_bytes)
            accounts.append(
                {
                    "index": i,
                    "address": acct.address,
                    "private_key_hex": priv_hex,
                }
            )
            del priv_bytes, acct, priv_hex  # clean up sensitive data
        return accounts

    def derive_private_key_hex_from_mnemonic(
        self,
        mnemonic: str,
        index: int,
        account: int = 0,
        change: Bip44Changes = Bip44Changes.CHAIN_EXT,
    ) -> str:
        """Derive the private key (hex) for a given account index from a mnemonic.

        Args:
            mnemonic (str): The plaintext BIP-39 mnemonic.
            index (int): Address index to derive.
            account (int): BIP-44 account index (default 0).
            change (Bip44Changes): Change chain (external/internal).

        Returns:
            str: Private key as a hex string (no 0x prefix).

        """
        seed_bytes = Bip39SeedGenerator(mnemonic).Generate()
        bip44_mst = Bip44.FromSeed(seed_bytes, Bip44Coins.ETHEREUM)
        # Build the context step by step to avoid long lines
        ctx = bip44_mst.Purpose().Coin().Account(account)
        ctx = ctx.Change(change)
        addr_ctx = ctx.AddressIndex(index)
        return addr_ctx.PrivateKey().Raw().ToHex()



from unittest.mock import patch

import pytest
from safe_eth.safe import SafeOperationEnum

from iwa.core.contracts.multisend import MultiSendCallOnlyContract, MultiSendContract


@pytest.fixture
def mock_contract_instance():
    with (
        patch("iwa.core.contracts.contract.ContractInstance.__init__", return_value=None),
        patch("iwa.core.contracts.contract.ContractInstance.call") as mock_call,
        patch("iwa.core.contracts.contract.ContractInstance.prepare_transaction") as mock_prep,
    ):
        yield mock_call, mock_prep


def test_encode_data():
    tx = {
        "operation": SafeOperationEnum.CALL,
        "to": "0x1111111111111111111111111111111111111111",
        "value": 100,
        "data": b"\x01\x02",
    }
    encoded = MultiSendCallOnlyContract.encode_data(tx)
    # Operation (1 byte) + To (20 bytes) + Value (32 bytes) + Data Length (32 bytes) + Data (2 bytes)
    # But wait, implementation uses HexBytes formatting which might produce different output if not careful.
    # Let's check length.
    # 1 + 20 + 32 + 32 + 2 = 87 bytes.
    assert len(encoded) == 87
    assert encoded[0] == 0  # CALL is 0


def test_to_bytes():
    tx1 = {
        "operation": SafeOperationEnum.CALL,
        "to": "0x1111111111111111111111111111111111111111",
        "value": 100,
        "data": b"\x01",
    }
    tx2 = {
        "operation": SafeOperationEnum.DELEGATE_CALL,
        "to": "0x2222222222222222222222222222222222222222",
        "value": 0,
        "data": b"",
    }
    encoded = MultiSendCallOnlyContract.to_bytes([tx1, tx2])
    # tx1: 1+20+32+32+1 = 86
    # tx2: 1+20+32+32+0 = 85
    # Total: 171
    assert len(encoded) == 171


def test_prepare_tx(mock_contract_instance):
    mock_call, mock_prep = mock_contract_instance
    mock_prep.return_value = {"data": "0x"}

    multisend = MultiSendCallOnlyContract("0xMulti", "gnosis")

    transactions = [
        {
            "operation": SafeOperationEnum.CALL,
            "to": "0x1111111111111111111111111111111111111111",
            "value": 100,
            "data": b"",
        }
    ]

    tx = multisend.prepare_tx("0xFrom", transactions)
    assert tx == {"data": "0x"}

    mock_prep.assert_called_once()
    call_args = mock_prep.call_args[1]
    assert call_args["method_name"] == "multiSend"
    assert "encoded_multisend_data" in call_args["method_kwargs"]
    assert call_args["tx_params"]["from"] == "0xFrom"
    assert call_args["tx_params"]["value"] == 100


def test_multisend_contract_init(mock_contract_instance):
    # Just verify it can be instantiated and has correct name
    ms = MultiSendContract("0xMulti", "gnosis")
    assert ms.name == "multisend"



"""Tests for the Web Server module."""

import pytest


class TestObscureUrl:
    """Tests for URL obscuring utility."""

    def test_obscure_url_hides_path(self):
        """Test that _obscure_url hides the path and query."""
        from iwa.web.server import _obscure_url

        result = _obscure_url("https://rpc.example.com/v1/api-key-12345")
        assert result == "https://rpc.example.com/..."
        assert "api-key" not in result

    def test_obscure_url_handles_empty(self):
        """Test that _obscure_url handles empty URLs."""
        from iwa.web.server import _obscure_url

        result = _obscure_url("")
        # Empty URL gives minimal output
        assert "..." in result

    def test_obscure_url_preserves_scheme_and_host(self):
        """Test that scheme and host are preserved."""
        from iwa.web.server import _obscure_url

        result = _obscure_url("wss://ws.alchemy.com/v2/secret")
        assert result == "wss://ws.alchemy.com/..."

    def test_obscure_url_handles_no_path(self):
        """Test URL with no path."""
        from iwa.web.server import _obscure_url

        result = _obscure_url("https://example.com")
        assert result == "https://example.com/..."


class TestInputValidation:
    """Tests for Pydantic input validation models."""

    def test_transaction_request_valid(self):
        """Test valid transaction request passes validation."""
        from iwa.web.server import TransactionRequest

        req = TransactionRequest(
            from_address="0x1234567890abcdef1234567890abcdef12345678",
            to_address="0xabcdef1234567890abcdef1234567890abcdef12",
            amount_eth=1.0,
            token="native",
            chain="gnosis",
        )
        assert req.from_address == "0x1234567890abcdef1234567890abcdef12345678"
        assert req.amount_eth == 1.0

    def test_transaction_request_valid_with_tag(self):
        """Test valid transaction request with tag passes validation."""
        from iwa.web.server import TransactionRequest

        req = TransactionRequest(
            from_address="my_wallet",
            to_address="receiver_wallet",
            amount_eth=1.0,
            token="native",
            chain="gnosis",
        )
        assert req.from_address == "my_wallet"

    def test_transaction_request_invalid_address_xss(self):
        """Test that XSS in address is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError) as exc:
            TransactionRequest(
                from_address="<script>alert(1)</script>",
                to_address="0x1234567890abcdef1234567890abcdef12345678",
                amount_eth=1.0,
                token="native",
                chain="gnosis",
            )
        assert "alphanumeric" in str(exc.value).lower()

    def test_transaction_request_invalid_address_format(self):
        """Test that invalid hex address is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError) as exc:
            TransactionRequest(
                from_address="0xINVALIDHEXADDRESS",
                to_address="0x1234567890abcdef1234567890abcdef12345678",
                amount_eth=1.0,
                token="native",
                chain="gnosis",
            )
        assert "invalid" in str(exc.value).lower()

    def test_transaction_request_empty_address(self):
        """Test that empty address is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError):
            TransactionRequest(
                from_address="",
                to_address="0x1234567890abcdef1234567890abcdef12345678",
                amount_eth=1.0,
                token="native",
                chain="gnosis",
            )

    def test_transaction_request_negative_amount(self):
        """Test that negative amounts are rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError) as exc:
            TransactionRequest(
                from_address="0x1234567890abcdef1234567890abcdef12345678",
                to_address="0xabcdef1234567890abcdef1234567890abcdef12",
                amount_eth=-1.0,
                token="native",
                chain="gnosis",
            )
        assert "positive" in str(exc.value).lower()

    def test_transaction_request_zero_amount(self):
        """Test that zero amount is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError):
            TransactionRequest(
                from_address="0x1234567890abcdef1234567890abcdef12345678",
                to_address="0xabcdef1234567890abcdef1234567890abcdef12",
                amount_eth=0,
                token="native",
                chain="gnosis",
            )

    def test_transaction_request_excessive_amount(self):
        """Test that excessive amount is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError) as exc:
            TransactionRequest(
                from_address="0x1234567890abcdef1234567890abcdef12345678",
                to_address="0xabcdef1234567890abcdef1234567890abcdef12",
                amount_eth=1e20,
                token="native",
                chain="gnosis",
            )
        assert "large" in str(exc.value).lower()

    def test_transaction_request_xss_chain(self):
        """Test that XSS in chain field is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import TransactionRequest

        with pytest.raises(ValidationError):
            TransactionRequest(
                from_address="0x1234567890abcdef1234567890abcdef12345678",
                to_address="0xabcdef1234567890abcdef1234567890abcdef12",
                amount_eth=1.0,
                token="native",
                chain="<script>",
            )

    def test_account_create_request_valid_tag(self):
        """Test valid tag passes validation."""
        from iwa.web.server import AccountCreateRequest

        req = AccountCreateRequest(tag="my_wallet_123")
        assert req.tag == "my_wallet_123"

    def test_account_create_request_none_tag(self):
        """Test None tag is allowed."""
        from iwa.web.server import AccountCreateRequest

        req = AccountCreateRequest(tag=None)
        assert req.tag is None

    def test_account_create_request_empty_tag(self):
        """Test empty tag becomes None."""
        from iwa.web.server import AccountCreateRequest

        req = AccountCreateRequest(tag="  ")
        assert req.tag is None

    def test_account_create_request_xss_tag(self):
        """Test that XSS in tag is rejected."""
        from pydantic import ValidationError

        from iwa.web.server import AccountCreateRequest

        with pytest.raises(ValidationError) as exc:
            AccountCreateRequest(tag="<script>alert(1)</script>")
        assert "alphanumeric" in str(exc.value).lower()

    def test_account_create_request_long_tag(self):
        """Test that too-long tags are rejected."""
        from pydantic import ValidationError

        from iwa.web.server import AccountCreateRequest

        with pytest.raises(ValidationError) as exc:
            AccountCreateRequest(tag="a" * 51)
        assert "long" in str(exc.value).lower()

    def test_account_create_request_with_underscore_hyphen(self):
        """Test that underscores and hyphens are allowed in tags."""
        from iwa.web.server import AccountCreateRequest

        req = AccountCreateRequest(tag="my-wallet_name")
        assert req.tag == "my-wallet_name"


class TestAppConfiguration:
    """Tests for app configuration."""

    def test_cors_middleware_configured(self):
        """Test that CORS middleware is configured."""
        from iwa.web.server import app

        # Check that user_middleware list is not empty (CORS was added)
        assert len(app.user_middleware) > 0

    def test_static_files_mounted(self):
        """Test that static files are mounted."""
        from iwa.web.server import app

        routes = [r.path for r in app.routes if hasattr(r, "path")]
        assert "/static" in routes



"""Multisend contract interaction."""

from typing import Dict, List, Optional, cast

from hexbytes import HexBytes
from safe_eth.safe import SafeOperationEnum

from iwa.core.constants import ABI_PATH
from iwa.core.contracts.contract import ContractInstance
from iwa.core.types import EthereumAddress

# MultiSend addresses (same across Ethereum, Base, Gnosis via Singleton Factory)
MULTISEND_CALL_ONLY_ADDRESS = EthereumAddress("0x40A2aCCbd92BCA938b02010E17A5b8929b49130D")
MULTISEND_ADDRESS = EthereumAddress("0xA238CBeb142c10Ef7Ad8442C6D1f9E89e07e7761")


class MultiSendCallOnlyContract(ContractInstance):
    """Class to interact with multisend (call only) contract."""

    name = "multisend_call_only"
    abi_path = ABI_PATH / "multisend_call_only.json"

    @staticmethod
    def encode_data(tx: Dict) -> bytes:
        """Encodes multisend transaction."""
        # Operation 1 byte
        operation = HexBytes("{:0>2x}".format(cast(SafeOperationEnum, tx.get("operation")).value))

        # Address 20 bytes
        to = HexBytes("{:0>40x}".format(int(cast(str, tx.get("to")), 16)))

        # Value 32 bytes
        value = HexBytes("{:0>64x}".format(cast(int, tx.get("value", 0))))

        # Data length 32 bytes
        data = cast(bytes, tx.get("data", b""))
        data_ = HexBytes(data)
        data_length = HexBytes("{:0>64x}".format(len(data_)))

        return operation + to + value + data_length + data_

    @staticmethod
    def to_bytes(multi_send_txs: List[Dict]) -> bytes:
        """Multi send tx list to bytes."""
        return b"".join([MultiSendCallOnlyContract.encode_data(tx) for tx in multi_send_txs])

    def prepare_tx(
        self,
        from_address: EthereumAddress,
        transactions: list,
    ) -> Optional[Dict]:
        """Prepare multisend transaction."""
        encoded_multisend_data = MultiSendCallOnlyContract.to_bytes(transactions)

        total_value_wei = sum([tx.get("value", 0) for tx in transactions])

        return self.prepare_transaction(
            method_name="multiSend",
            method_kwargs={"encoded_multisend_data": encoded_multisend_data},
            tx_params={
                "from": from_address,
                "value": total_value_wei,
            },
        )


class MultiSendContract(MultiSendCallOnlyContract):
    """Class to interact with multisend contract."""

    name = "multisend"
    abi_path = ABI_PATH / "multisend.json"



"""Script to verify ChainlistRPC functionality."""

from iwa.core.chainlist import ChainlistRPC


def main() -> None:
    """Run the verification script."""
    print("Initializing ChainlistRPC...")
    chainlist = ChainlistRPC()

    print("Fetching data...")
    chainlist.fetch_data()

    gnosis_chain_id = 100
    print(f"\n--- Gnosis Chain (ID: {gnosis_chain_id}) ---")

    all_rpcs = chainlist.get_rpcs(gnosis_chain_id)
    print(f"Total RPCs found: {len(all_rpcs)}")

    https_rpcs = chainlist.get_https_rpcs(gnosis_chain_id)
    print(f"HTTPS RPCs ({len(https_rpcs)}):")
    for url in https_rpcs[:5]:
        print(f"  - {url}")
    if len(https_rpcs) > 5:
        print("  ... and more")

    wss_rpcs = chainlist.get_wss_rpcs(gnosis_chain_id)
    print(f"WSS RPCs ({len(wss_rpcs)}):")
    for url in wss_rpcs[:5]:
        print(f"  - {url}")
    if len(wss_rpcs) > 5:
        print("  ... and more")

    print("\nTracking info for first 5 RPCs:")
    for node in all_rpcs[:5]:
        print(
            f"  - {node.url}: Tracking={node.is_tracking} (Privacy={node.privacy}, Tracking={node.tracking})"
        )


if __name__ == "__main__":
    main()



"""Lifecycle and operational flow tests for Olas ServiceManager."""

from unittest.mock import MagicMock, patch

import pytest

from iwa.plugins.olas.contracts.service import ServiceState
from iwa.plugins.olas.contracts.staking import StakingState
from iwa.plugins.olas.service_manager import ServiceManager


@pytest.fixture
def mock_wallet():
    """Mock wallet for testing."""
    wallet = MagicMock()
    wallet.master_account.address = "0x1234567890123456789012345678901234567890"
    wallet.key_storage.get_account.return_value.address = (
        "0xAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    )
    wallet.key_storage.generate_new_account.return_value.address = (
        "0xBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
    )
    wallet.sign_and_send_transaction.return_value = (True, {"status": 1})
    wallet.send.return_value = (True, {"status": 1})
    wallet.balance_service = MagicMock()
    wallet.transfer_service = MagicMock()
    # Default behavior for balance checking calls to avoid AttributeError
    wallet.balance_service.get_erc20_balance_wei.return_value = 10**20
    return wallet


@pytest.fixture
def mock_config():
    """Mock config for testing."""
    config = MagicMock()
    olas_config = MagicMock()
    olas_config.services = {}
    olas_config.get_service.return_value = None
    olas_config.get_service.return_value = None
    config.plugins = {"olas": olas_config}
    config.save_config = MagicMock()
    return config


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_create_service_success(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test successful service creation."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    # Setup Config with new OlasConfig structure
    mock_config_inst = mock_config_cls.return_value
    mock_olas_config = MagicMock()
    mock_olas_config.services = {}
    mock_olas_config.get_service.return_value = None
    mock_olas_config.get_service.return_value = None
    mock_config_inst.plugins = {"olas": mock_olas_config}
    mock_config_inst.save_config = MagicMock()

    # Setup Registry to return event
    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.extract_events.return_value = [
        {"name": "CreateService", "args": {"serviceId": 123}}
    ]

    # Setup Manager
    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    # Patch ChainInterfaces
    with patch("iwa.plugins.olas.service_manager.base.ChainInterfaces") as mock_chains:
        mock_chains.return_value.get.return_value.chain.get_token_address.return_value = None

        # Call create
        service_id = manager.create(bond_amount_wei=10)

        assert service_id == 123
        # Verify add_service was called
        mock_olas_config.add_service.assert_called_once()
        # Verify config was saved
        mock_config_inst.save_config.assert_called_once()


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_create_service_failures(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test service creation failure modes."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    mock_config_inst = mock_config_cls.return_value
    mock_olas_config = MagicMock()
    mock_olas_config.get_service.return_value = None
    mock_config_inst.plugins = {"olas": mock_olas_config}

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    # 1. Transaction fails
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    with patch("iwa.plugins.olas.service_manager.base.ChainInterfaces") as mock_chains:
        mock_chains.return_value.get.return_value.chain.get_token_address.return_value = None
        res = manager.create()
        assert res is None

    # Reset wallet
    mock_wallet.sign_and_send_transaction.return_value = (True, {})

    # 2. Event missing
    mock_registry_contract.return_value.extract_events.return_value = []
    with patch("iwa.plugins.olas.service_manager.base.ChainInterfaces") as mock_chains:
        mock_chains.return_value.get.return_value.chain.get_token_address.return_value = None
        res = manager.create()
        # Should effectively return None/False depending on implementation,
        # code says "if service_id is None: logger.error..." but proceeds to assign None
        # and returns service_id which is None.
        assert res is None

    # 3. Approval fails
    mock_registry_contract.return_value.extract_events.return_value = [
        {"name": "CreateService", "args": {"serviceId": 123}}
    ]
    # First tx (create) success, Second tx (approve) fails
    mock_wallet.sign_and_send_transaction.reset_mock()
    mock_wallet.sign_and_send_transaction.return_value = (
        None  # Clear return_value to let side_effect work
    )
    mock_wallet.sign_and_send_transaction.side_effect = [(True, {}), (False, {})]
    mock_wallet.transfer_service = MagicMock()
    mock_wallet.transfer_service.approve_erc20.return_value = False

    with patch("iwa.plugins.olas.service_manager.base.ChainInterfaces") as mock_chains:
        mock_chains.return_value.get.return_value.chain.get_token_address.return_value = (
            "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
        )
        res = manager.create(token_address_or_tag="0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f")
        assert res == 123 or res is False  # May succeed on create but fail on approval


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_create_service_with_approval(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test service creation with token approval."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    mock_config_inst = mock_config_cls.return_value
    mock_olas_config = MagicMock()
    mock_olas_config.get_service.return_value = None
    mock_config_inst.plugins = {"olas": mock_olas_config}

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.extract_events.return_value = [
        {"name": "CreateService", "args": {"serviceId": 123}}
    ]

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")
    mock_wallet.transfer_service = MagicMock()
    mock_wallet.transfer_service.approve_erc20.return_value = True

    with patch("iwa.plugins.olas.service_manager.base.ChainInterfaces") as mock_chains:
        mock_chains.return_value.get.return_value.chain.get_token_address.return_value = (
            "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
        )
        manager.create(token_address_or_tag="0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f")

        # Verify wallet transfer_service was used for approval
        assert mock_wallet.sign_and_send_transaction.call_count >= 1


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_activate_registration(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test service registration activation."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    mock_config_inst = mock_config_cls.return_value
    mock_olas_config = MagicMock()
    mock_service = MagicMock()
    mock_service.service_id = 123
    mock_service.chain_name = "gnosis"
    mock_service.token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_olas_config.get_service.return_value = mock_service
    mock_config_inst.plugins = {"olas": mock_olas_config}

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_registry_inst.extract_events.return_value = [{"name": "ActivateRegistration"}]

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    # Explicitly set mock return values to ensure they aren't overridden
    mock_wallet.balance_service.get_erc20_balance_wei.return_value = (
        50000000000000000000 * 2
    )  # Enough balance
    mock_wallet.transfer_service.get_erc20_allowance.return_value = (
        50000000000000000000 * 2
    )  # Enough allowance

    success = manager.activate_registration()
    assert success is True

    # Failures
    # 1. Wrong state
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.DEPLOYED,
        "security_deposit": 50000000000000000000,
    }
    assert manager.activate_registration() is False

    # 2. Tx fail
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    assert manager.activate_registration() is False

    # 3. Event missing
    mock_wallet.sign_and_send_transaction.return_value = (True, {})
    mock_registry_inst.extract_events.return_value = []
    assert manager.activate_registration() is False


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_register_agent(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test agent registration flow."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    mock_config_inst = mock_config_cls.return_value
    mock_olas_config = MagicMock()
    mock_service = MagicMock()
    mock_service.service_id = 123
    mock_service.chain_name = "gnosis"
    mock_service.token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_olas_config.get_service.return_value = mock_service
    mock_config_inst.plugins = {"olas": mock_olas_config}

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.ACTIVE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_registry_inst.extract_events.return_value = [{"name": "RegisterInstance"}]

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    success = manager.register_agent()
    assert success is True
    assert mock_olas_config.agent_address is not None

    # Failures
    # 1. Wrong state
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    assert manager.register_agent() is False

    # 2. Tx fail
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.ACTIVE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    assert manager.register_agent() is False

    # 3. Event missing
    mock_wallet.sign_and_send_transaction.return_value = (True, {})
    mock_registry_inst.extract_events.return_value = []
    assert manager.register_agent() is False


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_deploy(mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet):
    """Test service deployment."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    # Setup mock service
    mock_service = MagicMock()
    mock_service.service_id = 123
    mock_service.chain_name = "gnosis"
    mock_service.token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_service.multisig_address = None

    mock_olas_config = MagicMock()
    mock_olas_config.get_service.return_value = mock_service

    mock_config_inst = mock_config_cls.return_value
    mock_config_inst.plugins = {"olas": mock_olas_config}
    mock_config_inst.save_config = MagicMock()

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.FINISHED_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_registry_inst.extract_events.return_value = [
        {"name": "DeployService"},
        {
            "name": "CreateMultisigWithAgents",
            "args": {"multisig": "0xCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC"},
        },
    ]

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    multisig = manager.deploy()
    assert multisig == "0xCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC"
    mock_config_inst.save_config.assert_called()

    # Failures
    # 1. Wrong state
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    assert manager.deploy() is False  # returns False or None? Code says False if state wrong

    # 2. Tx fail
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.FINISHED_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    assert manager.deploy() is None

    # 3. Event missing (DeployService)
    mock_wallet.sign_and_send_transaction.return_value = (True, {})
    mock_registry_inst.extract_events.return_value = []
    assert manager.deploy() is None

    # 4. Multisig Missing
    mock_registry_inst.extract_events.return_value = [{"name": "DeployService"}]
    assert manager.deploy() is None


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch("iwa.plugins.olas.service_manager.base.ServiceManagerContract")
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_terminate(
    mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet
):
    """Test service termination."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    # Setup mock service
    mock_service = MagicMock()
    mock_service.service_id = 123
    mock_service.chain_name = "gnosis"
    mock_service.token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_service.staking_contract_address = None  # Not staked

    mock_olas_config = MagicMock()
    mock_olas_config.get_service.return_value = mock_service

    mock_config_inst = mock_config_cls.return_value
    mock_config_inst.plugins = {"olas": mock_olas_config}
    mock_config_inst.save_config = MagicMock()

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.DEPLOYED,
        "security_deposit": 50000000000000000000,
    }
    mock_registry_inst.extract_events.return_value = [{"name": "TerminateService"}]

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    success = manager.terminate()
    assert success is True

    # Failures
    # 1. Wrong state
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    assert manager.terminate() is False

    # 2. Staked
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.DEPLOYED,
        "security_deposit": 50000000000000000000,
    }
    manager.service.staking_contract_address = "0xStaked"
    assert manager.terminate() is False
    manager.service.staking_contract_address = None

    # 3. Tx fail
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    assert manager.terminate() is False

    # 4. No event
    mock_wallet.sign_and_send_transaction.return_value = (True, {})
    mock_registry_inst.extract_events.return_value = []
    assert manager.terminate() is False


@patch("iwa.plugins.olas.service_manager.base.Config")
@patch("iwa.plugins.olas.service_manager.base.ServiceRegistryContract")
@patch(
    "iwa.plugins.olas.service_manager.base.ServiceManagerContract"
)  # MUST mock specifically here
@patch("iwa.plugins.olas.service_manager.base.ContractCache")
def test_stake(mock_cache, mock_sm_contract, mock_registry_contract, mock_config_cls, mock_wallet):
    """Test service staking."""
    mock_cache.return_value.get_contract.side_effect = lambda cls, *a, **k: cls(*a, **k)
    # Setup mock service
    mock_service = MagicMock()
    mock_service.service_id = 123
    mock_service.chain_name = "gnosis"
    mock_service.token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_service.staking_contract_address = None

    mock_olas_config = MagicMock()
    mock_olas_config.get_service.return_value = mock_service

    mock_config_inst = mock_config_cls.return_value
    mock_config_inst.plugins = {"olas": mock_olas_config}
    mock_config_inst.save_config = MagicMock()

    mock_registry_inst = mock_registry_contract.return_value
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.DEPLOYED,
        "security_deposit": 50000000000000000000,
    }

    manager = ServiceManager(mock_wallet, service_key="gnosis:123")

    # Mock Staking Contract
    mock_staking = MagicMock()
    mock_staking.staking_token_address = "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"
    mock_staking.address = "0xDDdDddDdDdddDDddDDddDDDDdDdDDdDDdDDDDDDd"
    mock_staking.get_service_ids.return_value = []  # Not full
    mock_staking.max_num_services = 10
    mock_staking.min_staking_deposit = 100
    mock_staking.extract_events.return_value = [{"name": "ServiceStaked"}]
    mock_staking.get_staking_state.return_value = StakingState.STAKED
    mock_staking.get_requirements.return_value = {
        "staking_token": "0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f",
        "min_staking_deposit": 50000000000000000000,
        "num_agent_instances": 1,
        "required_agent_bond": 50000000000000000000,
    }

    success = manager.stake(mock_staking)
    assert success is True
    assert manager.service.staking_contract_address == "0xDDdDddDdDdddDDddDDddDDDDdDdDDdDDdDDDDDDd"
    mock_config_inst.save_config.assert_called()

    # Failures
    # 1. State not deployed
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.PRE_REGISTRATION,
        "security_deposit": 50000000000000000000,
    }
    assert manager.stake(mock_staking) is False

    # 2. Full
    mock_registry_inst.get_service.return_value = {
        "state": ServiceState.DEPLOYED,
        "security_deposit": 50000000000000000000,
    }
    mock_staking.get_service_ids.return_value = [0] * 10
    assert manager.stake(mock_staking) is False
    mock_staking.get_service_ids.return_value = []

    # 4. Approve fail
    mock_wallet.sign_and_send_transaction.return_value = (False, {})
    assert manager.stake(mock_staking) is False

    # 5. Stake fail
    # First tx (approve) success, second (stake) fail
    mock_wallet.sign_and_send_transaction.side_effect = [(True, {}), (False, {}), (False, {})]
    assert manager.stake(mock_staking) is False

    # 6. Event missing
    mock_wallet.sign_and_send_transaction.side_effect = [(True, {}), (True, {}), (True, {})]
    mock_staking.extract_events.return_value = []
    assert manager.stake(mock_staking) is False

    # 7. State check fail
    mock_wallet.sign_and_send_transaction.side_effect = [(True, {}), (True, {}), (True, {})]
    mock_staking.extract_events.return_value = [{"name": "ServiceStaked"}]
    mock_staking.get_staking_state.return_value = StakingState.NOT_STAKED
    assert manager.stake(mock_staking) is False



"""Test module."""

import asyncio

from iwa.core.wallet import Wallet
from iwa.plugins.olas.service_manager import ServiceManager

wallet = Wallet()


async def main():
    """Example of using CoW Swap on Gnosis Chain."""
    # await wallet.swap_tokens(
    #     account_address_or_tag="master",
    #     amount_eth=None,  # Swap entire balance
    #     sell_token_name="OLAS",
    #     buy_token_name="SDAI",
    #     chain_name="gnosis",
    #     fixed_buy_amount=False,
    # )

    service_manager = ServiceManager(wallet=wallet)
    service_manager.create()


if __name__ == "__main__":  # pragma: no cover
    asyncio.run(main())



"""Olas General Router."""

from fastapi import APIRouter, Depends
from loguru import logger

from iwa.web.dependencies import verify_auth

router = APIRouter(tags=["olas"])


@router.get(
    "/price",
    summary="Get OLAS Price",
    description="Get the current price of OLAS token in EUR from CoinGecko.",
)
def get_olas_price(auth: bool = Depends(verify_auth)):
    """Get current OLAS token price in EUR from CoinGecko."""
    try:
        from iwa.core.pricing import PriceService

        price_service = PriceService()
        price = price_service.get_token_price("autonolas", "eur")

        return {"price_eur": price, "symbol": "OLAS"}
    except Exception as e:
        logger.error(f"Error fetching OLAS price: {e}")
        return {"price_eur": None, "symbol": "OLAS", "error": str(e)}



"""Tests for Wallets Refactor."""

from unittest.mock import MagicMock, patch

import pytest

from iwa.tui.screens.wallets import WalletsScreen


@pytest.fixture
def mock_wallet():
    """Mock wallet."""
    return MagicMock()


def test_wallets_screen_initialization(mock_wallet):
    """Test that WalletsScreen can be initialized with the new import structure."""
    with patch("iwa.core.chain.ChainInterfaces"):
        # Mock MonitorWorker import if needed, but it should be imported from workers.py now
        screen = WalletsScreen(wallet=mock_wallet)
        assert screen is not None
        assert hasattr(screen, "monitor_worker") or hasattr(screen, "start_monitor")


def test_monitor_worker_integration():
    """Test that MonitorWorker is imported from the correct location."""
    from iwa.tui.screens.wallets import MonitorWorker as WalletsMonitorWorker
    from iwa.tui.workers import MonitorWorker as DefinedMonitorWorker

    assert WalletsMonitorWorker is DefinedMonitorWorker



"""Gnosis Safe plugin."""

from typing import Dict, Optional

import typer

from iwa.core.keys import KeyStorage
from iwa.core.plugins import Plugin


class GnosisPlugin(Plugin):
    """Gnosis Safe Plugin."""

    @property
    def name(self) -> str:
        """Get plugin name."""
        return "gnosis"

    def get_cli_commands(self) -> Dict[str, callable]:
        """Get CLI commands."""
        return {"create-safe": self.create_safe_command}

    def create_safe_command(
        self,
        tag: Optional[str] = typer.Option(
            None,
            "--tag",
            "-t",
            help="Tag for this account",
        ),
        owners: str = typer.Option(
            ...,
            "--owners",
            "-o",
            help="Comma-separated list of owner addresses or tags.",
        ),
        threshold: int = typer.Option(
            ...,
            "--threshold",
            "-h",
            help="Number of required confirmations.",
        ),
        chain_name: str = typer.Option(
            "gnosis",
            "--chain",
            "-c",
            help="Chain to deploy the multisig on.",
        ),
    ):
        """Create a new multisig account (Safe)"""
        from iwa.core.services import AccountService, SafeService

        key_storage = KeyStorage()
        account_service = AccountService(key_storage)
        safe_service = SafeService(key_storage, account_service)

        owner_list = [owner.strip() for owner in owners.split(",")]
        try:
            safe_service.create_safe(
                deployer_tag_or_address="master",
                owner_tags_or_addresses=owner_list,
                threshold=threshold,
                chain_name=chain_name,
                tag=tag,
            )
        except ValueError as e:
            typer.echo(f"Error: {e}")
            raise typer.Exit(code=1) from e



"""Tests for Wallet module."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from iwa.core.wallet import Wallet


@pytest.fixture
def mock_keys_and_services():
    """Mock keys and services."""
    with (
        patch("iwa.core.wallet.KeyStorage") as mock_ks,
        patch("iwa.core.wallet.AccountService") as mock_as,
        patch("iwa.core.wallet.BalanceService") as mock_bs,
        patch("iwa.core.wallet.SafeService") as mock_ss,
        patch("iwa.core.wallet.TransactionService") as mock_ts,
        patch("iwa.core.wallet.TransferService") as mock_trs,
        patch("iwa.core.wallet.PluginService") as mock_ps,
        patch("iwa.core.wallet.init_db") as mock_init_db,
        patch("iwa.core.wallet.configure_logger"),
    ):
        yield {
            "key_storage": mock_ks,
            "account_service": mock_as,
            "balance_service": mock_bs,
            "safe_service": mock_ss,
            "transaction_service": mock_ts,
            "transfer_service": mock_trs,
            "plugin_service": mock_ps,
            "init_db": mock_init_db,
        }


@pytest.fixture
def wallet(mock_keys_and_services):
    """Wallet fixture."""
    return Wallet()


def test_init(mock_keys_and_services):
    """Test initialization."""
    wallet = Wallet()
    assert wallet.key_storage == mock_keys_and_services["key_storage"].return_value
    mock_keys_and_services["init_db"].assert_called_once()


def test_master_account(wallet, mock_keys_and_services):
    """Test master account property."""
    # Accesses property on account_service instance


def test_get_token_address(wallet, mock_keys_and_services):
    """Test get_token_address."""
    wallet.get_token_address("OLAS", "gnosis")
    mock_keys_and_services["account_service"].return_value.get_token_address.assert_called_with(
        "OLAS", "gnosis"
    )


def test_get_accounts_balances(wallet, mock_keys_and_services):
    """Test get_accounts_balances."""
    # Mock account data
    mock_keys_and_services["account_service"].return_value.get_account_data.return_value = {
        "0x1": {"tag": "one"},
        "0x2": {"tag": "two"},
    }

    # Mock balance service
    mock_bs = mock_keys_and_services["balance_service"].return_value
    mock_bs.get_native_balance_eth.return_value = 1.0
    mock_bs.get_erc20_balance_eth.return_value = 2.0

    # Test with no token names
    data, balances = wallet.get_accounts_balances("gnosis")
    assert data == {"0x1": {"tag": "one"}, "0x2": {"tag": "two"}}
    assert balances is None

    # Test with token names
    # Mock ThreadPoolExecutor to run synchronously or just return futures
    # Since we can't easily suppress the real ThreadPoolExecutor context manager used in the code without patching it,
    # let's patch it in the test function scope.

    with patch("iwa.core.wallet.ThreadPoolExecutor") as mock_executor:
        # returns context manager
        mock_context = MagicMock()
        mock_executor.return_value.__enter__.return_value = mock_context

        # mock submit to return a future
        mock_future1 = MagicMock()
        mock_future1.result.return_value = ("0x1", "native", 1.0)

        mock_future2 = MagicMock()
        mock_future2.result.return_value = ("0x1", "OLAS", 2.0)

        mock_context.submit.side_effect = [
            mock_future1,
            mock_future2,
            mock_future1,
            mock_future2,
        ]  # Just cycling mocks

        # We need to rely on what the code does: it iterates over accounts, then tokens.
        # 2 accounts * 2 tokens = 4 calls.

        # Simpler approach: let the real ThreadPoolExecutor run but mock the balance service methods which are already mocked.
        # The issue is the code uses `fetch_balance` inner function.
        pass


# Re-implementing test_get_accounts_balances with delegation verification via patching ThreadPoolExecutor
# effectively mocking concurrency.


def test_get_accounts_balances_concurrency(wallet, mock_keys_and_services):
    """Test get_accounts_balances concurrency."""
    mock_keys_and_services["account_service"].return_value.get_account_data.return_value = {
        "0x1": {"tag": "one"}
    }

    with patch("iwa.core.wallet.ThreadPoolExecutor") as mock_executor:
        mock_context = MagicMock()
        mock_executor.return_value.__enter__.return_value = mock_context

        mock_future_native = MagicMock()
        mock_future_native.result.return_value = ("0x1", "native", 1.5)

        mock_future_token = MagicMock()
        mock_future_token.result.return_value = ("0x1", "OLAS", 10.0)

        # The loop order in wallet.py: for addr in accounts: for t in tokens: submit
        # so for 1 account and 2 tokens: native, OLAS
        mock_context.submit.side_effect = [mock_future_native, mock_future_token]

        accounts, balances = wallet.get_accounts_balances("gnosis", ["native", "OLAS"])

        assert accounts == {"0x1": {"tag": "one"}}
        # balances structure: {addr: {token: val}}
        assert balances["0x1"]["native"] == 1.5
        assert balances["0x1"]["OLAS"] == 10.0


def test_send_native_transfer(wallet, mock_keys_and_services):
    """Test send_native_transfer."""
    mock_keys_and_services["transfer_service"].return_value.send.return_value = "0xhash"
    success, tx_hash = wallet.send_native_transfer("0xfrom", "0xto", 100, "gnosis")
    assert success is True
    assert tx_hash == "0xhash"
    mock_keys_and_services["transfer_service"].return_value.send.assert_called_with(
        from_address_or_tag="0xfrom",
        to_address_or_tag="0xto",
        amount_wei=100,
        token_address_or_name="native",
        chain_name="gnosis",
    )


def test_sign_and_send_transaction(wallet, mock_keys_and_services):
    """Test sign_and_send_transaction."""
    wallet.sign_and_send_transaction({"to": "0x1"}, "owner", "gnosis")
    mock_keys_and_services["transaction_service"].return_value.sign_and_send.assert_called_with(
        {"to": "0x1"}, "owner", "gnosis", None
    )


def test_send_erc20_transfer(wallet, mock_keys_and_services):
    """Test send_erc20_transfer."""
    mock_keys_and_services["transfer_service"].return_value.send.return_value = "0xhash"
    success, tx_hash = wallet.send_erc20_transfer("0xfrom", "0xto", 100, "0xtoken", "gnosis")
    assert success is True
    assert tx_hash == "0xhash"
    mock_keys_and_services["transfer_service"].return_value.send.assert_called_with(
        from_address_or_tag="0xfrom",
        to_address_or_tag="0xto",
        amount_wei=100,
        token_address_or_name="0xtoken",
        chain_name="gnosis",
    )


def test_send(wallet, mock_keys_and_services):
    """Test send."""
    wallet.send("0xfrom", "0xto", 100)
    mock_keys_and_services["transfer_service"].return_value.send.assert_called_with(
        "0xfrom", "0xto", 100, "native", "gnosis"
    )


def test_multi_send(wallet, mock_keys_and_services):
    """Test multi_send."""
    txs = [{"to": "0x1", "value": 1}]
    wallet.multi_send("0xfrom", txs, "gnosis")
    mock_keys_and_services["transfer_service"].return_value.multi_send.assert_called_with(
        "0xfrom", txs, "gnosis"
    )


def test_balances_getters(wallet, mock_keys_and_services):
    """Test balances getters."""
    mock_bs = mock_keys_and_services["balance_service"].return_value

    wallet.get_native_balance_eth("0x1", "gnosis")
    mock_bs.get_native_balance_eth.assert_called_with("0x1", "gnosis")

    wallet.get_native_balance_wei("0x1", "gnosis")
    mock_bs.get_native_balance_wei.assert_called_with("0x1", "gnosis")

    wallet.get_erc20_balance_eth("0x1", "OLAS", "gnosis")
    mock_bs.get_erc20_balance_eth.assert_called_with("0x1", "OLAS", "gnosis")

    wallet.get_erc20_balance_wei("0x1", "OLAS", "gnosis")
    mock_bs.get_erc20_balance_wei.assert_called_with("0x1", "OLAS", "gnosis")


def test_erc20_operations(wallet, mock_keys_and_services):
    """Test erc20 operations."""
    mock_trs = mock_keys_and_services["transfer_service"].return_value

    wallet.get_erc20_allowance("owner", "spender", "token", "gnosis")
    mock_trs.get_erc20_allowance.assert_called_with("owner", "spender", "token", "gnosis")

    wallet.approve_erc20("owner", "spender", "token", 100, "gnosis")
    mock_trs.approve_erc20.assert_called_with("owner", "spender", "token", 100, "gnosis")

    wallet.transfer_from_erc20("from", "sender", "recipient", "token", 100, "gnosis")
    mock_trs.transfer_from_erc20.assert_called_with(
        "from", "sender", "recipient", "token", 100, "gnosis"
    )


@pytest.mark.asyncio
async def test_swap(wallet, mock_keys_and_services):
    """Test swap."""
    mock_trs = mock_keys_and_services["transfer_service"].return_value

    # Mock swap as async method
    mock_trs.swap = AsyncMock(return_value=True)

    result = await wallet.swap("account", 1.0, "SELL", "BUY", "gnosis")
    assert result is True

    # Check if await matches call args
    # Note: Enum handling might need import, passing string/value might be tested
    args, kwargs = mock_trs.swap.call_args
    assert args[0] == "account"
    assert args[1] == 1.0


def test_drain(wallet, mock_keys_and_services):
    """Test drain."""
    wallet.drain("from", "to", "gnosis")
    mock_keys_and_services["transfer_service"].return_value.drain.assert_called_with(
        "from", "to", "gnosis"
    )



"""Tests for PluginService."""

from unittest.mock import MagicMock, patch

from iwa.core.plugins import Plugin
from iwa.core.services.plugin import PluginService


class MockPlugin(Plugin):
    """Mock plugin for testing."""

    @property
    def name(self):
        return "mock_plugin"

    def get_cli_commands(self):
        return {"mock": lambda: None}


def test_plugin_service_init():
    """Test PluginService initialization loads plugins."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()
        assert service.plugins_package == "iwa.plugins"


def test_discover_plugins_import_error():
    """Test _discover_plugins handles ImportError."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()

    with patch(
        "iwa.core.services.plugin.importlib.import_module",
        side_effect=ImportError("Module not found"),
    ):
        plugins = service._discover_plugins()
        assert plugins == []


def test_discover_plugins_no_path():
    """Test _discover_plugins handles package without __path__."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()

    with patch("iwa.core.services.plugin.importlib.import_module") as mock_import:
        mock_package = MagicMock(spec=[])  # No __path__
        del mock_package.__path__  # Ensure hasattr returns False
        mock_import.return_value = mock_package

        plugins = service._discover_plugins()
        assert plugins == []


def test_load_plugins_module_error():
    """Test _load_plugins handles module import errors."""
    with (
        patch.object(PluginService, "_discover_plugins", return_value=["bad_module"]),
        patch(
            "iwa.core.services.plugin.importlib.import_module",
            side_effect=ImportError("Bad module"),
        ),
    ):
        # Should not raise, just log error
        service = PluginService()

        assert "bad_module" not in service.loaded_plugins


def test_get_plugin():
    """Test get_plugin returns correct plugin."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()
        mock_plugin = MockPlugin()
        service.loaded_plugins["mock_plugin"] = mock_plugin

        result = service.get_plugin("mock_plugin")

        assert result == mock_plugin


def test_get_plugin_not_found():
    """Test get_plugin returns None for unknown plugin."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()

        result = service.get_plugin("nonexistent")

        assert result is None


def test_get_all_plugins():
    """Test get_all_plugins returns copy of plugins."""
    with patch.object(PluginService, "_load_plugins"):
        service = PluginService()
        mock_plugin = MockPlugin()
        service.loaded_plugins["mock_plugin"] = mock_plugin

        result = service.get_all_plugins()

        assert "mock_plugin" in result
        # Verify it's a copy
        result["new_plugin"] = None
        assert "new_plugin" not in service.loaded_plugins


def test_skip_already_loaded():
    """Test _load_plugins skips already loaded plugins."""
    with patch.object(PluginService, "_discover_plugins", return_value=["mock"]):
        with patch.object(PluginService, "_load_plugins"):
            service = PluginService()

        # Pre-populate loaded plugins with key matching discovered name
        mock_plugin = MockPlugin()
        service.loaded_plugins["mock"] = mock_plugin

        # Now load - should skip "mock" since it's in loaded_plugins
        with patch("iwa.core.services.plugin.importlib.import_module") as mock_import:
            service._load_plugins()
            mock_import.assert_not_called()



"""Tests for SafeTransactionExecutor."""

from unittest.mock import MagicMock, patch

import pytest
from safe_eth.safe.safe_tx import SafeTx

from iwa.core.services.safe_executor import (
    MIN_SIGNATURE_LENGTH,
    SAFE_TX_STATS,
    SafeTransactionExecutor,
)


@pytest.fixture(autouse=True)
def reset_stats():
    """Reset SAFE_TX_STATS before each test to prevent state leakage."""
    for key in SAFE_TX_STATS:
        SAFE_TX_STATS[key] = 0
    yield


@pytest.fixture
def mock_chain_interface():
    ci = MagicMock()
    ci.current_rpc = "http://mock-rpc"
    ci.DEFAULT_MAX_RETRIES = 6
    ci._is_rate_limit_error.return_value = False
    ci._is_connection_error.return_value = False
    ci._handle_rpc_error.return_value = {"should_retry": True}
    return ci


@pytest.fixture
def executor(mock_chain_interface):
    return SafeTransactionExecutor(mock_chain_interface)


@pytest.fixture
def mock_safe_tx():
    """Mock SafeTx with valid 65-byte signature."""
    tx = MagicMock(spec=SafeTx)
    tx.safe_tx_gas = 100000
    tx.base_gas = 0
    tx.gas_price = 1000000000
    tx.to = "0xTo"
    tx.value = 0
    tx.data = b""
    tx.operation = 0
    # Valid signatures must be >= 65 bytes (one ECDSA signature)
    tx.signatures = b"x" * 65
    return tx


@pytest.fixture
def mock_safe():
    s = MagicMock()
    s.estimate_tx_gas.return_value = 100000
    s.retrieve_nonce.return_value = 5
    return s


# =============================================================================
# Test: Basic execution success
# =============================================================================


def test_execute_success_first_try(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test successful execution on first attempt."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        mock_safe_tx.execute.return_value = b"tx_hash"

        success, tx_hash, receipt = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is True
        assert tx_hash == "0x" + b"tx_hash".hex()
        assert mock_safe_tx.execute.call_count == 1


# =============================================================================
# Test: Tuple vs bytes return handling
# =============================================================================


def test_execute_handles_tuple_return(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test that executor handles safe_tx.execute() returning tuple (tx_hash, tx)."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        # Simulate tuple return: (tx_hash_bytes, tx_data)
        mock_safe_tx.execute.return_value = (b"tx_hash", {"gas": 21000})

        success, tx_hash, receipt = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is True
        assert tx_hash == "0x" + b"tx_hash".hex()


def test_execute_handles_bytes_return(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test that executor handles safe_tx.execute() returning raw bytes."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        mock_safe_tx.execute.return_value = b"raw_hash"

        success, tx_hash, receipt = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is True
        assert tx_hash.startswith("0x")


def test_execute_handles_string_return(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test that executor handles safe_tx.execute() returning hex string."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        mock_safe_tx.execute.return_value = "0xabcdef1234567890"

        success, tx_hash, receipt = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is True
        assert tx_hash == "0xabcdef1234567890"


# =============================================================================
# Test: Signature validation
# =============================================================================


def test_execute_fails_on_empty_signatures(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Verify we fail immediately if no signatures exist."""
    mock_safe_tx.signatures = b""  # Empty signatures

    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        with patch("time.sleep"):
            success, error, _ = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

    assert success is False
    assert "No valid signatures" in error
    assert mock_safe_tx.execute.call_count == 0  # Never tried to execute


def test_execute_fails_on_truncated_signatures(
    executor, mock_chain_interface, mock_safe_tx, mock_safe
):
    """Verify we detect signatures shorter than 65 bytes."""
    mock_safe_tx.signatures = b"x" * 30  # Too short (need 65)

    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        with patch("time.sleep"):
            success, error, _ = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

    assert success is False
    assert "No valid signatures" in error or str(MIN_SIGNATURE_LENGTH) in error


def test_execute_fails_on_none_signatures(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Verify we handle None signatures gracefully."""
    mock_safe_tx.signatures = None

    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        with patch("time.sleep"):
            success, error, _ = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

    assert success is False
    assert "No valid signatures" in error


# =============================================================================
# Test: Error classification (GS0xx codes)
# =============================================================================


def test_gs020_fails_fast(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """GS020 (signatures too short) should not trigger retries."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.call.side_effect = ValueError("execution reverted: GS020")

        with patch("time.sleep"):
            success, error, _ = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is False
        assert "GS020" in error
        assert mock_safe_tx.execute.call_count == 0  # Never got to execute


def test_gs026_fails_fast(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """GS026 (invalid owner) should not trigger retries."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.call.side_effect = ValueError("execution reverted: GS026")

        with patch("time.sleep"):
            success, error, _ = executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert success is False
        assert "GS026" in error


@pytest.mark.parametrize(
    "error_code,is_signature_error",
    [
        ("GS020", True),  # Signatures data too short
        ("GS021", True),  # Invalid signature data pointer
        ("GS024", True),  # Invalid contract signature
        ("GS026", True),  # Invalid owner
        ("GS025", False),  # Invalid nonce (not a signature error)
        ("GS010", False),  # Not enough gas
        ("GS013", False),  # Safe transaction failed
    ],
)
def test_error_classification(executor, error_code, is_signature_error):
    """Verify correct classification of Safe error codes."""
    error = ValueError(f"execution reverted: {error_code}")
    result = executor._is_signature_error(error)
    assert result == is_signature_error


# =============================================================================
# Test: Retry behavior
# =============================================================================


def test_retry_on_transient_error(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test that transient errors trigger retries without modifying tx."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = [ConnectionError("Network timeout"), b"success_hash"]
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        mock_chain_interface._is_connection_error.return_value = True

        with patch("time.sleep"):
            success, tx_hash, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is True
        assert mock_safe_tx.execute.call_count == 2
        # Gas should NOT have changed (we don't modify after signing)
        assert mock_safe_tx.safe_tx_gas == 100000


def test_retry_on_nonce_error(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test nonce refresh on GS025 error."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = [ValueError("GS025: invalid nonce"), b"success_hash"]
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )

        new_tx = MagicMock(spec=SafeTx)
        new_tx.signatures = b"x" * 65
        executor._refresh_nonce = MagicMock(return_value=new_tx)

        with patch("time.sleep"):
            executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert executor._refresh_nonce.called
        assert new_tx.execute.called


def test_retry_on_rpc_error(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test RPC rotation on rate limit error."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = [ValueError("Rate limit exceeded"), b"success_hash"]
        mock_chain_interface._is_rate_limit_error.return_value = True
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )

        with patch("time.sleep"):
            executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert mock_chain_interface._handle_rpc_error.called
        assert mock_chain_interface._handle_rpc_error.call_count == 1


def test_fail_after_max_retries(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test failure after exhausting all retries."""
    executor.max_retries = 2
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = ValueError("Persistent error")

        with patch("time.sleep"):
            success, tx_hash_or_err, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is False
        assert mock_safe_tx.execute.call_count == 3  # 1 initial + 2 retries


# =============================================================================
# Test: State preservation during retries
# =============================================================================


def test_retry_preserves_signatures_despite_clearing(
    executor, mock_chain_interface, mock_safe_tx, mock_safe
):
    """Verify that retries don't corrupt/lose signatures even if library clears them."""
    original_signatures = mock_safe_tx.signatures

    # Define a side effect that clears signatures on success (mimicking safe-eth-py)
    def execute_side_effect(key, **kwargs):
        # Simulate library behavior: clears signatures after "executing"
        mock_safe_tx.signatures = b""
        return b"hash"

    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        # Scenario:
        # 1. Execute success (sigs cleared) but Receipt not found (triggering retry)
        # 2. Retry: Execute called again (must have restored sigs) -> Success -> Receipt found

        mock_chain_interface.web3.eth.wait_for_transaction_receipt.side_effect = [
            ValueError("Transaction not found"),
            MagicMock(status=1),
        ]

        mock_safe_tx.execute.side_effect = execute_side_effect
        mock_chain_interface._is_connection_error.return_value = False

        with patch("time.sleep"):
            success, tx_hash, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is True
        # Signatures should be restored after the loop (or at least valid during 2nd call)
        # We assert they match original at the end because the finally block restores if changed?
        # Wait, finally restores IF signatures != backup.
        # If call 2 succeeds, execute() sets signatures to b"" AGAIN at the end of call 2.
        # So at the end of execution, signatures ARE empty locally if we updated them?
        # NO, the finally block runs AFTER safe_tx.execute returns.
        # So after call 2 returns (sigs=b""), finally restores them (sigs=original).
        # So they should be original.
        assert mock_safe_tx.signatures == original_signatures
        assert mock_safe_tx.execute.call_count == 2


def test_retry_preserves_gas(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Verify that retries don't modify safe_tx_gas (which would invalidate signatures)."""
    original_gas = mock_safe_tx.safe_tx_gas

    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = [ConnectionError("timeout"), b"hash"]
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        mock_chain_interface._is_connection_error.return_value = True

        with patch("time.sleep"):
            executor.execute_with_retry("0xSafe", mock_safe_tx, ["key1"])

        assert mock_safe_tx.safe_tx_gas == original_gas


# =============================================================================
# Test: Gas estimation (_estimate_safe_tx_gas)
# =============================================================================


def test_estimate_safe_tx_gas_with_buffer(executor, mock_safe):
    """Test gas estimation applies buffer correctly."""
    mock_safe.estimate_tx_gas.return_value = 100_000
    mock_safe_tx = MagicMock()
    mock_safe_tx.to = "0xDest"
    mock_safe_tx.value = 0
    mock_safe_tx.data = b""
    mock_safe_tx.operation = 0

    result = executor._estimate_safe_tx_gas(mock_safe, mock_safe_tx)

    # Default buffer is 1.5, so 100000 * 1.5 = 150000
    assert result == 150_000


def test_estimate_safe_tx_gas_caps_at_10x(executor, mock_safe):
    """Test gas estimation respects x10 cap when base_estimate is provided."""
    mock_safe.estimate_tx_gas.return_value = 500_000  # High estimate
    mock_safe_tx = MagicMock()
    mock_safe_tx.to = "0xDest"
    mock_safe_tx.value = 0
    mock_safe_tx.data = b""
    mock_safe_tx.operation = 0

    # 500000 * 1.5 = 750000, but base_estimate * 10 = 50000
    result = executor._estimate_safe_tx_gas(mock_safe, mock_safe_tx, base_estimate=5_000)

    # Should be capped at 5000 * 10 = 50000
    assert result == 50_000


def test_estimate_safe_tx_gas_fallback_on_failure(executor, mock_safe):
    """Test gas estimation uses fallback when estimation fails."""
    mock_safe.estimate_tx_gas.side_effect = Exception("Estimation failed")
    mock_safe_tx = MagicMock()
    mock_safe_tx.to = "0xDest"
    mock_safe_tx.value = 0
    mock_safe_tx.data = b""
    mock_safe_tx.operation = 0

    result = executor._estimate_safe_tx_gas(mock_safe, mock_safe_tx)

    assert result == executor.DEFAULT_FALLBACK_GAS


# =============================================================================
# Test: Error decoding (_decode_revert_reason)
# =============================================================================


def test_decode_revert_reason_with_hex_data(executor):
    """Test decoding when error contains hex data."""
    # Create an error with hex data that might be decodable
    error = ValueError("execution reverted: 0x08c379a0...")

    with patch("iwa.core.services.safe_executor.ErrorDecoder") as mock_decoder:
        mock_decoder.return_value.decode.return_value = [("Error", "Insufficient balance", "ERC20")]
        result = executor._decode_revert_reason(error)

    # Note: Due to hex matching, this should find the data and attempt decode
    assert result == "Insufficient balance (from ERC20)"


def test_decode_revert_reason_no_hex_data(executor):
    """Test decoding when error has no hex data."""
    error = ValueError("Some generic error without hex")

    result = executor._decode_revert_reason(error)

    assert result is None


def test_decode_revert_reason_decode_fails(executor):
    """Test decoding when decoder returns None."""
    error = ValueError("error: 0xdeadbeef")

    with patch("iwa.core.services.safe_executor.ErrorDecoder") as mock_decoder:
        mock_decoder.return_value.decode.return_value = None
        result = executor._decode_revert_reason(error)

    assert result is None


# =============================================================================
# Test: Error classification
# =============================================================================


def test_classify_error_gas_error(executor):
    """Test classification of gas-related errors."""
    error = ValueError("intrinsic gas too low")
    result = executor._classify_error(error)

    assert result["is_gas_error"] is True
    assert result["is_nonce_error"] is False


def test_classify_error_revert(executor):
    """Test classification of revert errors."""
    error = ValueError("execution reverted: some reason")
    result = executor._classify_error(error)

    assert result["is_revert"] is True


def test_classify_error_out_of_gas(executor):
    """Test classification of out of gas errors."""
    error = ValueError("out of gas")
    result = executor._classify_error(error)

    assert result["is_gas_error"] is True


# =============================================================================
# Test: Transaction failures
# =============================================================================


def test_transaction_reverts_onchain(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test handling when transaction is mined but reverts (status 0)."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.return_value = b"tx_hash"
        # Receipt with status 0 (reverted)
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=0
        )

        with patch("time.sleep"):
            success, error, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is False
        assert "reverted" in error.lower()


def test_check_receipt_status_dict_format(executor):
    """Test receipt status check with dict-style receipt."""
    # Dict-style receipt (not MagicMock)
    receipt_dict = {"status": 1, "gasUsed": 21000}
    assert executor._check_receipt_status(receipt_dict) is True

    receipt_dict_failed = {"status": 0}
    assert executor._check_receipt_status(receipt_dict_failed) is False


def test_simulation_revert_not_nonce(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test handling when simulation reverts with non-nonce error."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        # Simulation fails with generic revert
        mock_safe_tx.call.side_effect = ValueError("execution reverted: insufficient funds")

        with patch("time.sleep"):
            success, error, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is False
        assert "insufficient funds" in error.lower() or "reverted" in error.lower()


def test_gas_error_strategy_triggers_retry(
    executor, mock_chain_interface, mock_safe_tx, mock_safe
):
    """Test that gas errors trigger retry with gas increase strategy."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = [
            ValueError("intrinsic gas too low"),
            b"tx_hash",
        ]
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )

        with patch("time.sleep"):
            success, tx_hash, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        # Should have retried and succeeded
        assert success is True
        assert mock_safe_tx.execute.call_count == 2


def test_rpc_rotation_stops_when_should_not_retry(
    executor, mock_chain_interface, mock_safe_tx, mock_safe
):
    """Test that execution stops when RPC handler says not to retry."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        mock_safe_tx.execute.side_effect = ValueError("Rate limit exceeded")
        mock_chain_interface._is_rate_limit_error.return_value = True
        mock_chain_interface._handle_rpc_error.return_value = {"should_retry": False}

        with patch("time.sleep"):
            success, error, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is False
        # Only 1 attempt because should_retry=False
        assert mock_safe_tx.execute.call_count == 1


# =============================================================================
# Test: Fee bumping on base fee errors
# =============================================================================


def test_fee_error_triggers_bump(executor, mock_chain_interface, mock_safe_tx, mock_safe):
    """Test that fee errors trigger gas price bump on retry."""
    with patch.object(executor, "_recreate_safe_client", return_value=mock_safe):
        # First attempt fails with fee error, second succeeds
        mock_safe_tx.execute.side_effect = [
            ValueError("max fee per gas less than block base fee: maxFeePerGas: 596, baseFee: 681"),
            b"tx_hash",
        ]
        mock_chain_interface.web3.eth.wait_for_transaction_receipt.return_value = MagicMock(
            status=1
        )
        # Mock fee calculation
        mock_chain_interface.web3.eth.get_block.return_value = {"baseFeePerGas": 700}
        mock_chain_interface.web3.eth.max_priority_fee = 1

        with patch("time.sleep"):
            success, tx_hash, receipt = executor.execute_with_retry(
                "0xSafe", mock_safe_tx, ["key1"]
            )

        assert success is True
        assert mock_safe_tx.execute.call_count == 2
        # Second call should have tx_gas_price (bumped), not eip1559_speed
        second_call_kwargs = mock_safe_tx.execute.call_args_list[1][1]
        assert "tx_gas_price" in second_call_kwargs


def test_fee_error_classification(executor):
    """Test classification of fee-related errors."""
    fee_errors = [
        "max fee per gas less than block base fee",
        "transaction underpriced",
        "maxFeePerGas too low",
        "fee too low for mempool",
    ]
    for error_msg in fee_errors:
        error = ValueError(error_msg)
        result = executor._classify_error(error)
        assert result["is_fee_error"] is True, f"Should detect fee error: {error_msg}"


def test_calculate_bumped_gas_price_eip1559(executor, mock_chain_interface):
    """Test bumped gas price calculation for EIP-1559 chains."""
    mock_chain_interface.web3.eth.get_block.return_value = {"baseFeePerGas": 1000}
    mock_chain_interface.web3.eth.max_priority_fee = 10

    # With 1.3x bump factor: base_fee * 1.3 * 1.5 + priority = 1000 * 1.3 * 1.5 + 10 = 1960
    result = executor._calculate_bumped_gas_price(1.3)

    assert result is not None
    assert result == int(1000 * 1.3 * 1.5) + 10


def test_calculate_bumped_gas_price_legacy(executor, mock_chain_interface):
    """Test bumped gas price calculation for legacy chains."""
    mock_chain_interface.web3.eth.get_block.return_value = {}  # No baseFeePerGas
    mock_chain_interface.web3.eth.gas_price = 2000

    # Legacy: gas_price * bump_factor = 2000 * 1.3 = 2600
    result = executor._calculate_bumped_gas_price(1.3)

    assert result is not None
    assert result == int(2000 * 1.3)



"""ERC20 contract interaction."""

from typing import Dict, Optional

from iwa.core.constants import ABI_PATH
from iwa.core.contracts.contract import ContractInstance
from iwa.core.types import EthereumAddress


class ERC20Contract(ContractInstance):
    """Class to interact with ERC20 contracts."""

    name = "erc20"
    abi_path = ABI_PATH / "erc20.json"

    def __init__(self, address: EthereumAddress, chain_name: str = "gnosis"):
        """Initialize ERC20 contract instance."""
        super().__init__(address, chain_name)

        self.decimals = self.call("decimals")
        self.symbol = self.call("symbol")
        self.name = self.call("name")
        self.total_supply = self.call("totalSupply")

    def allowance_wei(self, owner: EthereumAddress, spender: EthereumAddress) -> int:
        """Allowance"""
        return self.call("allowance", owner, spender)

    def allowance_eth(self, owner: EthereumAddress, spender: EthereumAddress) -> float:
        """Allowance in human readable format"""
        return self.allowance_wei(owner, spender) / (10**self.decimals)

    def balance_of_wei(self, account: EthereumAddress) -> int:
        """Balance of"""
        return self.call("balanceOf", account)

    def balance_of_eth(self, account: EthereumAddress) -> float:
        """Balance of in human readable format"""
        return self.balance_of_wei(account) / (10**self.decimals)

    def prepare_transfer_tx(
        self,
        from_address: EthereumAddress,
        to: EthereumAddress,
        amount_wei: int,
    ) -> Optional[Dict]:
        """Transfer."""
        return self.prepare_transaction(
            method_name="transfer",
            method_kwargs={"to": to, "amount": amount_wei},
            tx_params={"from": from_address},
        )

    def prepare_transfer_from_tx(
        self,
        from_address: EthereumAddress,
        sender: EthereumAddress,
        recipient: EthereumAddress,
        amount_wei: int,
    ) -> Optional[Dict]:
        """Transfer from."""
        return self.prepare_transaction(
            method_name="transferFrom",
            method_kwargs={"_sender": sender, "_recipient": recipient, "_amount": amount_wei},
            tx_params={"from": from_address},
        )

    def prepare_approve_tx(
        self,
        from_address: EthereumAddress,
        spender: EthereumAddress,
        amount_wei: int,
    ) -> Optional[Dict]:
        """Approve."""
        return self.prepare_transaction(
            method_name="approve",
            method_kwargs={"spender": spender, "amount": amount_wei},
            tx_params={"from": from_address},
        )



"""Plugin Service module."""

import importlib
import inspect
import pkgutil
from typing import Dict, List, Optional

from loguru import logger

from iwa.core.plugins import Plugin


class PluginService:
    """Manages plugin discovery, loading, and lifecycle."""

    def __init__(self, plugins_package: str = "iwa.plugins"):
        """Initialize PluginService.

        Args:
            plugins_package: Python package path to search for plugins.

        """
        self.plugins_package = plugins_package
        self.loaded_plugins: Dict[str, Plugin] = {}
        self._load_plugins()

    def _discover_plugins(self) -> List[str]:
        """Discover available plugins in the plugins package."""
        try:
            package = importlib.import_module(self.plugins_package)
            if not hasattr(package, "__path__"):
                return []

            return [name for _, name, is_pkg in pkgutil.iter_modules(package.__path__) if is_pkg]
        except ImportError:
            logger.warning(f"Could not import plugins package: {self.plugins_package}")
            return []

    def _load_plugins(self) -> None:
        """Load all discovered plugins."""
        from iwa.core.models import Config

        plugin_names = self._discover_plugins()
        config = Config()

        for name in plugin_names:
            if name in self.loaded_plugins:
                continue

            try:
                module_name = f"{self.plugins_package}.{name}"
                module = importlib.import_module(module_name)

                # Find Plugin subclass
                for _, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and issubclass(obj, Plugin) and obj is not Plugin:
                        try:
                            plugin_instance = obj()
                            # Verify unique name
                            if plugin_instance.name in self.loaded_plugins:
                                logger.warning(
                                    f"Plugin name collision: {plugin_instance.name}. Skipping."
                                )
                                continue

                            # Register plugin's config model if it has one
                            if plugin_instance.config_model:
                                config.register_plugin_config(
                                    plugin_instance.name, plugin_instance.config_model
                                )

                            self.loaded_plugins[plugin_instance.name] = plugin_instance
                            plugin_instance.on_load()
                            logger.info(f"Loaded plugin: {plugin_instance.name}")
                        except Exception as e:
                            logger.error(f"Failed to instantiate plugin {name}: {e}")

            except Exception as e:
                logger.error(f"Failed to load plugin module {name}: {e}")

    def get_plugin(self, name: str) -> Optional[Plugin]:
        """Get a loaded plugin by name."""
        return self.loaded_plugins.get(name)

    def get_all_plugins(self) -> Dict[str, Plugin]:
        """Get all loaded plugins."""
        # Return a copy to prevent modification
        return self.loaded_plugins.copy()



from unittest.mock import MagicMock, patch

import pytest
from textual.widgets import Input, Select

from iwa.tui.app import IwaApp
from iwa.tui.screens.wallets import WalletsScreen


@pytest.fixture
def mock_wallet():
    with patch("iwa.tui.app.Wallet") as mock_wallet_cls:
        wallet = mock_wallet_cls.return_value
        wallet.key_storage.accounts = {
            "0x1": MagicMock(address="0x1", tag="Acc1"),
            "0x2": MagicMock(address="0x2", tag="Acc2"),
        }
        wallet.account_service = MagicMock()
        wallet.account_service.accounts = wallet.key_storage.accounts
        wallet.account_service.get_account_data.side_effect = (
            lambda: wallet.account_service.accounts
        )
        wallet.balance_service = MagicMock()

        # Helper to make resolve_account work with key_storage for consistency
        wallet.account_service.resolve_account.side_effect = (
            lambda tag: wallet.key_storage.get_account(tag)
        )

        wallet.get_native_balance_eth.return_value = 0.0
        wallet.get_erc20_balance_eth.return_value = 0.0
        wallet.send.return_value = "0xHash"

        # Mock PluginService
        wallet.plugin_service = MagicMock()
        wallet.plugin_service.get_all_plugins.return_value = {}
        yield wallet


@pytest.fixture(autouse=True)
def mock_deps():
    with (
        patch("iwa.tui.screens.wallets.EventMonitor"),
        patch("iwa.tui.screens.wallets.PriceService") as mock_price,
        patch("iwa.core.db.SentTransaction"),
        patch("iwa.core.db.log_transaction"),
        patch("iwa.tui.screens.wallets.ChainInterfaces") as mock_chains,
    ):
        mock_price.return_value.get_token_price.return_value = 10.0

        # Setup Chain Interface Mock
        mock_interface = MagicMock()
        mock_interface.tokens = {"TOKEN": "0xToken"}
        mock_interface.chain.native_currency = "ETH"
        mock_chains.return_value.get.return_value = mock_interface
        mock_chains.return_value.items.return_value = [("gnosis", mock_interface)]

        yield {"chains": mock_chains}


@pytest.mark.asyncio
async def test_fetch_balances_flow(mock_wallet, mock_deps):
    app = IwaApp()
    # Patch call_from_thread
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)):
        view = app.query_one(WalletsScreen)

        # Configure returns on both legacy and service
        mock_wallet.get_native_balance_eth.return_value = 1.2345
        mock_wallet.balance_service.get_native_balance_eth.return_value = 1.2345
        mock_wallet.balance_service.get_erc20_balance_with_retry.return_value = 500.0

        # Trigger fetch directly
        view.balance_cache = {}

        # Trigger (call impl directly to avoid threading issues)
        view.chain_token_states["gnosis"].add("TOKEN")
        # In the refactored view, we call fetch_all_balances
        # We'll wait for the worker to finish
        worker = view.fetch_all_balances(view.active_chain, ["TOKEN"])
        await worker.wait()

        # Verify calls made (can check either legacy or service depending on how it's implemented)
        mock_wallet.balance_service.get_native_balance_eth.assert_called()
        mock_wallet.balance_service.get_erc20_balance_with_retry.assert_called()

        # Verify cache state
        assert view.balance_cache["gnosis"]["0x1"]["NATIVE"] == "1.2345"
        assert view.balance_cache["gnosis"]["0x1"]["TOKEN"] == "500.0000"


@pytest.mark.asyncio
async def test_chain_changed(mock_wallet, mock_deps):
    app = IwaApp()
    # Patch call_from_thread
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)) as pilot:
        view = app.query_one(WalletsScreen)

        # Select different chain
        select = app.query_one("#chain_select")
        select.value = "ethereum"
        await pilot.pause(1.0)

        assert view.active_chain == "ethereum"

        # Test Invalid chain (no RPC)
        chains = mock_deps["chains"]
        chains.return_value.get.return_value.chain.rpc = ""
        select.value = "base"
        await pilot.pause()


@pytest.mark.asyncio
async def test_send_transaction_coverage(mock_wallet, mock_deps):
    app = IwaApp()
    # Patch call_from_thread
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(200, 200)) as pilot:
        view = app.query_one(WalletsScreen)
        # Force table height
        app.query_one("#accounts_table").styles.height = 10
        # Wait for workers to populate
        # In the new design, monitor_workers are created in start_monitor
        if not view.monitor_workers:
            view.start_monitor()
        assert len(view.monitor_workers) > 0
        await pilot.pause()

        # Test validation failures
        # 1. No from
        app.query_one("#from_addr", Select).value = Select.BLANK
        mock_wallet.send.reset_mock()
        btn = app.query_one("#send_btn")
        btn.focus()
        await pilot.press("enter")
        # Assert not sent
        mock_wallet.send.assert_not_called()

        # 2. No to
        app.query_one("#from_addr", Select).value = "0x1"
        app.query_one("#to_addr", Select).value = Select.BLANK
        btn = app.query_one("#send_btn")
        btn.focus()
        await pilot.press("enter")
        mock_wallet.send.assert_not_called()

        # 3. No amount
        app.query_one("#to_addr", Select).value = "0x2"
        app.query_one("#amount", Input).value = ""
        btn = app.query_one("#send_btn")
        btn.focus()
        await pilot.press("enter")
        mock_wallet.send.assert_not_called()

        # 4. Valid Send NATIVE
        app.query_one("#amount", Input).value = "1.0"
        app.query_one("#token", Select).value = "native"
        mock_wallet.send.return_value = "0xTxHash"

        # Call worker directly
        view.send_tx_worker("0x1", "0x2", "native", 1.0)
        await pilot.pause()

        # Verify wallet.send called
        mock_wallet.send.assert_called()

        # 5. Valid Send ERC20
        mock_wallet.send.reset_mock()
        view.send_tx_worker("0x1", "0x2", "TOKEN", 1.0)
        await pilot.pause()
        mock_wallet.send.assert_called()


@pytest.mark.asyncio
async def test_watchdog_logic(mock_wallet, mock_deps):
    app = IwaApp()
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)):
        view = app.query_one(WalletsScreen)

        # 1. Test "Everything Loaded" -> No Retry
        view.balance_cache = {
            "gnosis": {
                "0x1": {"NATIVE": "1.0", "TOKEN": "10.0"},
                "0x2": {"NATIVE": "2.0", "TOKEN": "20.0"},
            }
        }
        view.chain_token_states["gnosis"] = {"TOKEN"}

        # Should NOT trigger fetch
        with patch.object(view, "fetch_all_balances") as mock_fetch:
            view.check_balance_loading_status("gnosis")
            mock_fetch.assert_not_called()

        # 2. Test "Missing Native" -> Retry
        view.balance_cache["gnosis"]["0x1"]["NATIVE"] = "Loading..."
        with patch.object(view, "fetch_all_balances") as mock_fetch:
            view.check_balance_loading_status("gnosis")
            mock_fetch.assert_called_with("gnosis", ["TOKEN"])

        # 3. Test "Missing Chain in Cache" -> Retry
        del view.balance_cache["gnosis"]
        with patch.object(view, "fetch_all_balances") as mock_fetch:
            view.check_balance_loading_status("gnosis")
            mock_fetch.assert_called_with("gnosis", ["TOKEN"])

        # Restore for next
        view.balance_cache = {"gnosis": {}}

        # 4. Test "Missing Address in Cache" -> Retry
        view.balance_cache["gnosis"] = {}  # Empty
        with patch.object(view, "fetch_all_balances") as mock_fetch:
            view.check_balance_loading_status("gnosis")
            mock_fetch.assert_called_with("gnosis", ["TOKEN"])


@pytest.mark.asyncio
async def test_monitor_handler(mock_wallet, mock_deps):
    app = IwaApp()
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)):
        view = app.query_one(WalletsScreen)

        # Simulate txs
        txs = [
            {
                "hash": "0xHash1",
                "timestamp": 1700000000,
                "from": "0x1",
                "to": "0x2",
                "token": "NATIVE",
                "value": 10**18,
                "chain": "gnosis",
            },
            {
                "hash": "0xHash2",
                "timestamp": None,
                "from": "0x3",
                "to": "0x4",
                "token": "DAI",
                "value": 500,
                "chain": "gnosis",
            },
        ]

        view.handle_new_txs(txs)

        # Verify table rows added
        table = app.query_one("#tx_table")
        assert table.row_count >= 2
        # Verify first row details
        assert "Detected" in str(table.get_row("0xHash1"))


@pytest.mark.asyncio
async def test_token_fetch_retry_and_failure(mock_wallet, mock_deps):
    app = IwaApp()
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)):
        view = app.query_one(WalletsScreen)
        view.balance_cache = {}
        view.chain_token_states["gnosis"] = {"TOKEN"}

        # Check worker logic
        worker = view.monitor_workers[0]
        # We can't easily check internal state of the worker loop without complex mocking
        # assert worker.monitor.chain_name in ["gnosis", "ethereum", "base"]

        # Limit to 1 account to control call count
        mock_wallet.key_storage.accounts = {"0x1": MagicMock(address="0x1", tag="Acc1")}

        # Patch time.sleep to avoid wait
        with patch("time.sleep"):
            # Case 1: Success
            mock_wallet.balance_service.get_erc20_balance_with_retry.return_value = 100.0

            worker = view.fetch_all_balances("gnosis", ["TOKEN"])
            await worker.wait()
            # verify call made
            mock_wallet.balance_service.get_erc20_balance_with_retry.assert_called()
            # Should have updated cache (4 decimals)
            assert view.balance_cache["gnosis"]["0x1"]["TOKEN"] == "100.0000"


@pytest.mark.asyncio
async def test_send_transaction_failure(mock_wallet, mock_deps):
    app = IwaApp()
    app.call_from_thread = lambda cb, *args, **kwargs: cb(*args, **kwargs)

    async with app.run_test(size=(160, 80)):
        view = app.query_one(WalletsScreen)
        mock_wallet.send.side_effect = Exception("Tx Failed")

        # Just ensure it doesn't crash
        view.send_tx_worker("0x1", "0x2", "native", 1.0)


# --- Tests migrated from test_tui_clipboard_chain.py ---


@pytest.mark.asyncio
async def test_wallets_view_clipboard(mock_wallet, mock_deps):
    """Test clipboard functionality for account and transaction cells."""
    app = IwaApp()
    async with app.run_test() as _:
        view = app.query_one(WalletsScreen)

        # Test on_account_cell_selected
        mock_event = MagicMock()
        mock_event.coordinate.column = 1
        mock_event.value = "0xAddress"

        with patch.dict("sys.modules", {"pyperclip": MagicMock()}):
            view.on_account_cell_selected(mock_event)

        # Test on_tx_cell_selected
        mock_event_tx = MagicMock()
        mock_event_tx.coordinate.column = 0
        mock_event_tx.data_table.columns.values.return_value = [
            MagicMock(label="Hash"),
            MagicMock(label="Status"),
        ]
        mock_event_tx.cell_key.row_key.value = "0xFullHash"

        with patch.dict("sys.modules", {"pyperclip": MagicMock()}):
            view.on_tx_cell_selected(mock_event_tx)


from textual.widget import Widget


class DummySelect(Widget):
    """Dummy Select widget for testing chain change."""

    def __init__(self, *args, **kwargs):
        super().__init__(id=kwargs.get("id"))

    def set_options(self, options):
        pass


@pytest.mark.asyncio
async def test_wallets_view_chain_change_detailed(mock_wallet, mock_deps):
    """Test chain change with detailed mocking."""
    app = IwaApp()
    async with app.run_test() as _:
        view = app.query_one(WalletsScreen)

        # Mock ChainInterfaces
        with patch("iwa.tui.screens.wallets.ChainInterfaces") as mock_chains:
            mock_interface = MagicMock()
            mock_interface.chain.rpc = "http://rpc"
            mock_interface.chain.native_currency = "ETH"
            mock_interface.tokens = {"TOKEN": "0xToken"}
            mock_chains.return_value.get.return_value = mock_interface

            # Patch Select with DummyWidget to satisfy isinstance(w, Widget)
            with patch("iwa.tui.screens.wallets.Select", side_effect=DummySelect):
                # Chain changed event
                mock_event = MagicMock()
                mock_event.value = "gnosis"
                mock_event.control.value = "gnosis"
                view.active_chain = "ethereum"

                # Trigger
                await view.on_chain_changed(mock_event)

                assert view.active_chain == "gnosis"


# --- Tests migrated from test_keystorage_edge_cases.py ---


@pytest.mark.asyncio
async def test_wallets_view_actions():
    """Test WalletsScreen action methods."""
    with patch("iwa.core.db.db") as mock_db:
        mock_db.is_closed.return_value = True

        app = IwaApp()
        async with app.run_test() as _:
            view = app.query_one(WalletsScreen)

            # Test action_refresh
            with patch.object(view, "refresh_accounts") as mock_refresh:
                view.action_refresh()
                mock_refresh.assert_called_with(force=True)

            # Test on_unmount
            view.start_monitor()
            with patch.object(view, "stop_monitor") as mock_stop:
                view.on_unmount()
                mock_stop.assert_called()

            # Test monitor_callback
            with patch.object(view, "handle_new_txs") as _:
                with patch.object(app, "call_from_thread") as mock_call:
                    view.monitor_callback([])
                    mock_call.assert_called_with(view.handle_new_txs, [])


@pytest.mark.asyncio
async def test_wallets_view_resolve_tag(mock_wallet, mock_deps):
    """Test resolve_tag method."""
    view = WalletsScreen(mock_wallet)

    mock_acc = MagicMock()
    mock_acc.address = "0xAddress"
    mock_acc.tag = "MyTag"

    mock_accounts = MagicMock()
    mock_accounts.values.return_value = [mock_acc]

    mock_wallet.key_storage.accounts = mock_accounts
    mock_wallet.account_service.accounts = mock_accounts

    tag = view.resolve_tag("0xAddress")
    assert tag == "MyTag"

    # Test fallback
    mock_accounts.values.return_value = []
    tag = view.resolve_tag("0xAddress")
    assert tag == "0xAddr...ress"


@pytest.mark.asyncio
async def test_create_safe_modal_cancel():
    """Test CreateSafeModal cancel button."""
    from iwa.tui.modals import CreateSafeModal

    modal = CreateSafeModal([])

    mock_event = MagicMock()
    mock_event.button.id = "cancel"
    with patch.object(modal, "dismiss") as mock_dismiss:
        modal.on_button_pressed(mock_event)
        mock_dismiss.assert_called()


# --- Tests migrated from test_core_db_integration.py ---

from unittest.mock import PropertyMock

from textual.widgets import DataTable


@pytest.mark.asyncio
async def test_create_safe_worker_no_rpc(mock_wallet, mock_deps):
    """Test create_safe_worker when no RPC is configured."""
    view = WalletsScreen(mock_wallet)

    with patch.object(WalletsScreen, "app", new_callable=PropertyMock) as mock_app_prop:
        mock_app = MagicMock()
        mock_app_prop.return_value = mock_app
        view.notify = MagicMock()

        with patch("iwa.tui.screens.wallets.ChainInterfaces") as mock_chains:
            mock_interface = MagicMock()
            mock_interface.chain.rpc = None
            mock_chains.return_value.get.return_value = mock_interface

            view.create_safe_worker("Tag", 1, ["0x1"], ["gnosis"])

            assert mock_app.call_from_thread.call_count >= 1


@pytest.mark.asyncio
async def test_create_safe_worker_exception(mock_wallet, mock_deps):
    """Test create_safe_worker handles exceptions."""
    view = WalletsScreen(mock_wallet)

    with patch.object(WalletsScreen, "app", new_callable=PropertyMock) as mock_app_prop:
        mock_app = MagicMock()
        mock_app_prop.return_value = mock_app
        view.notify = MagicMock()

        with patch("iwa.tui.screens.wallets.ChainInterfaces") as mock_chains:
            mock_interface = MagicMock()
            mock_interface.chain.rpc = "http://rpc"
            mock_chains.return_value.get.return_value = mock_interface

            mock_wallet.key_storage.create_safe.side_effect = Exception("Create Failed")

            view.create_safe_worker("Tag", 1, ["0x1"], ["gnosis"])

            assert mock_app.call_from_thread.call_count >= 1


@pytest.mark.asyncio
async def test_action_quit():
    """Test IwaApp action_quit."""
    app = IwaApp()
    with patch.object(app, "exit") as mock_exit:
        await app.action_quit()
        mock_exit.assert_called_once()


@pytest.mark.asyncio
async def test_wallets_view_lifecycle(mock_wallet, mock_deps):
    """Test WalletsScreen lifecycle (on_mount/compose)."""
    app = IwaApp()
    async with app.run_test() as _:
        view = app.query_one(WalletsScreen)
        assert view is not None
        table = view.query_one("#accounts_table", DataTable)
        assert len(table.columns) > 0


@pytest.mark.asyncio
async def test_wallets_view_copy_address_fallback(mock_wallet, mock_deps):
    """Test clipboard fallback when pyperclip fails."""
    app = IwaApp()
    async with app.run_test() as _:
        view = app.query_one(WalletsScreen)

        mock_event = MagicMock()
        mock_event.coordinate.column = 1
        mock_event.value = "0xAddr"

        mock_pyperclip = MagicMock()
        mock_pyperclip.copy.side_effect = Exception("No clipboard")

        with patch.dict("sys.modules", {"pyperclip": mock_pyperclip}):
            with patch("iwa.tui.app.IwaApp.copy_to_clipboard") as mock_copy:
                view.on_account_cell_selected(mock_event)
                mock_copy.assert_called_with("0xAddr")


@pytest.mark.asyncio
async def test_enrich_logs_api_failure(mock_wallet, mock_deps):
    """Test enrich_and_log_txs handles API failures."""
    app = IwaApp()
    async with app.run_test():
        view = app.query_one(WalletsScreen)

        txs = [{"hash": "0x1", "token": "TOKEN", "chain": "gnosis"}]

        with patch("iwa.tui.screens.wallets.PriceService") as mock_price:
            mock_price.return_value.get_token_price.return_value = None

            with patch("iwa.core.db.log_transaction") as _:
                with patch("iwa.tui.screens.wallets.ChainInterfaces"):
                    if not view.monitor_workers:
                        view.start_monitor()
                    view.enrich_and_log_txs(txs)
                    # Just verify it doesn't crash



"""Tests for Olas Plugin."""

from unittest.mock import MagicMock, patch

from iwa.plugins.olas.models import OlasConfig
from iwa.plugins.olas.plugin import OlasPlugin


class TestOlasPlugin:
    """Tests for OlasPlugin class."""

    def test_name_property(self):
        """Test plugin name property."""
        plugin = OlasPlugin()
        assert plugin.name == "olas"

    def test_config_model_returns_olas_config(self):
        """Test config_model property returns OlasConfig."""
        plugin = OlasPlugin()
        assert plugin.config_model == OlasConfig

    def test_get_cli_commands(self):
        """Test get_cli_commands returns dict with commands."""
        plugin = OlasPlugin()
        commands = plugin.get_cli_commands()

        assert isinstance(commands, dict)
        assert "create" in commands
        assert callable(commands["create"])

    @patch("iwa.plugins.olas.plugin.Wallet")
    @patch("iwa.plugins.olas.plugin.ServiceManager")
    def test_create_service(self, mock_sm_class, mock_wallet_class):
        """Test create_service calls ServiceManager.create."""
        mock_wallet = MagicMock()
        mock_wallet_class.return_value = mock_wallet

        mock_manager = MagicMock()
        mock_sm_class.return_value = mock_manager

        plugin = OlasPlugin()
        plugin.create_service(
            chain_name="gnosis",
            owner="0x1234",
            token="OLAS",
            bond=100,
        )

        mock_wallet_class.assert_called_once()
        mock_sm_class.assert_called_once_with(mock_wallet)
        mock_manager.create.assert_called_once_with(
            chain_name="gnosis",
            service_owner_address_or_tag="0x1234",
            token_address_or_tag="OLAS",
            bond_amount_wei=100,
        )

    @patch("iwa.plugins.olas.plugin.Wallet")
    @patch("iwa.plugins.olas.plugin.ServiceManager")
    def test_create_service_defaults(self, mock_sm_class, mock_wallet_class):
        """Test create_service with default parameters."""
        mock_wallet = MagicMock()
        mock_wallet_class.return_value = mock_wallet

        mock_manager = MagicMock()
        mock_sm_class.return_value = mock_manager

        plugin = OlasPlugin()
        # Note: when called directly (not via typer), defaults are OptionInfo objects
        # so we can only verify the method was called
        plugin.create_service()

        mock_wallet_class.assert_called_once()
        mock_sm_class.assert_called_once_with(mock_wallet)
        mock_manager.create.assert_called_once()  # Any arguments



import sys
from unittest.mock import MagicMock, patch

import pytest
from typer.testing import CliRunner


@pytest.fixture
def iwa_cli_module():
    mock_cowpy = MagicMock()
    modules_to_patch = {
        "cowdao_cowpy": mock_cowpy,
        "cowdao_cowpy.common": MagicMock(),
        "cowdao_cowpy.common.chains": MagicMock(),
        "cowdao_cowpy.app_data": MagicMock(),
        "cowdao_cowpy.app_data.utils": MagicMock(),
        "cowdao_cowpy.contracts": MagicMock(),
        "cowdao_cowpy.contracts.order": MagicMock(),
        "cowdao_cowpy.contracts.sign": MagicMock(),
        "cowdao_cowpy.cow": MagicMock(),
        "cowdao_cowpy.cow.swap": MagicMock(),
        "cowdao_cowpy.order_book": MagicMock(),
        "cowdao_cowpy.order_book.api": MagicMock(),
        "cowdao_cowpy.order_book.config": MagicMock(),
        "cowdao_cowpy.order_book.generated": MagicMock(),
        "cowdao_cowpy.order_book.generated.model": MagicMock(),
    }

    with patch.dict(sys.modules, modules_to_patch):
        if "iwa.core.cli" in sys.modules:
            del sys.modules["iwa.core.cli"]
        if "iwa.core.wallet" in sys.modules:
            pass

        with patch("iwa.core.wallet.Wallet"):
            import iwa.core.cli

            yield iwa.core.cli.iwa_cli


runner = CliRunner()


@pytest.fixture
def cli(iwa_cli_module):
    return iwa_cli_module


@pytest.fixture
def mock_key_storage():
    with patch("iwa.core.cli.KeyStorage") as mock:
        yield mock.return_value


@pytest.fixture
def mock_wallet():
    with patch("iwa.core.cli.Wallet") as mock:
        yield mock.return_value


def test_account_create(cli, mock_key_storage):
    result = runner.invoke(cli, ["wallet", "create", "--tag", "test"])
    assert result.exit_code == 0
    mock_key_storage.generate_new_account.assert_called_with("test")


def test_account_create_error(cli, mock_key_storage):
    mock_key_storage.generate_new_account.side_effect = ValueError("Error creating account")
    result = runner.invoke(cli, ["wallet", "create", "--tag", "test"])
    assert result.exit_code == 1
    assert "Error: Error creating account" in result.stdout


def test_account_list(cli, mock_wallet):
    mock_wallet.get_accounts_balances.return_value = ({}, None)
    with (
        patch("iwa.core.cli.list_accounts") as mock_list_accounts,
        patch("iwa.core.cli.ChainInterfaces"),
    ):
        result = runner.invoke(cli, ["wallet", "list", "--chain", "gnosis", "--balances", "native"])
        assert result.exit_code == 0
        mock_wallet.get_accounts_balances.assert_called_with("gnosis", ["native"])
        mock_list_accounts.assert_called_once()


def test_account_send(cli, mock_wallet):
    result = runner.invoke(
        cli, ["wallet", "send", "--from", "sender", "--to", "receiver", "--amount", "1.0"]
    )
    assert result.exit_code == 0
    mock_wallet.send.assert_called()


def test_erc20_transfer_from(cli, mock_wallet):
    result = runner.invoke(
        cli,
        [
            "wallet",
            "transfer-from",
            "--from",
            "from",
            "--sender",
            "sender",
            "--recipient",
            "recipient",
            "--token",
            "token",
            "--amount",
            "1.0",
        ],
    )
    assert result.exit_code == 0
    mock_wallet.transfer_from_erc20.assert_called()


def test_erc20_approve(cli, mock_wallet):
    result = runner.invoke(
        cli,
        [
            "wallet",
            "approve",
            "--owner",
            "owner",
            "--spender",
            "spender",
            "--token",
            "token",
            "--amount",
            "1.0",
        ],
    )
    assert result.exit_code == 0
    mock_wallet.approve_erc20.assert_called()


def test_drain_wallet(cli, mock_wallet):
    result = runner.invoke(cli, ["wallet", "drain", "--from", "from", "--to", "to"])
    assert result.exit_code == 0
    mock_wallet.drain.assert_called()



from unittest.mock import MagicMock, patch

import pytest

from iwa.core.contracts.cache import ContractCache
from iwa.plugins.olas.contracts.staking import StakingContract


@pytest.fixture
def mock_chain_interface():
    with patch("iwa.core.contracts.contract.ChainInterfaces") as mock_chains:
        mock_interface = MagicMock()
        mock_chains.return_value.get.return_value = mock_interface

        # Mock web3 and contract
        mock_web3_backend = MagicMock()
        mock_interface.web3._web3 = mock_web3_backend

        mock_contract = MagicMock()
        mock_web3_backend.eth.contract.return_value = mock_contract

        # Mock with_retry to execute the function
        mock_interface.with_retry.side_effect = lambda func, **kwargs: func()

        # Yield both interface and contract mock
        yield mock_interface, mock_contract


def test_staking_contract_lazy_loading(mock_chain_interface):
    """Verify StakingContract init does NOT make RPC calls."""
    mock_interface, mock_contract = mock_chain_interface

    # Reset ContractCache
    ContractCache().clear()

    # Instantiate logic
    contract = StakingContract(address="0x123", chain_name="gnosis")

    # Assert NO calls to call yet (since we mock with_retry to execute immediately, 0 calls means 0 executions)
    assert mock_interface.with_retry.call_count == 0

    # Setup return value
    # livenessPeriod is a property that calls "livenessPeriod" on contract
    mock_contract.functions.livenessPeriod.return_value.call.return_value = 3600

    # Access property
    val = contract.liveness_period
    assert val == 3600

    # Assert 1 call
    assert mock_contract.functions.livenessPeriod.return_value.call.call_count == 1

    # Access again
    val = contract.liveness_period

    # Assert still 1 call (cached)
    assert mock_contract.functions.livenessPeriod.return_value.call.call_count == 1


def test_contract_cache_singleton(mock_chain_interface):
    """Verify ContractCache returns same instance and reuses property cache."""
    mock_interface, mock_contract = mock_chain_interface
    ContractCache().clear()

    c1 = ContractCache().get_contract(StakingContract, "0xABC", "gnosis")
    c2 = ContractCache().get_contract(StakingContract, "0xabc", "gnosis")  # Check ignore case

    assert c1 is c2

    # populate cache on c1
    mock_contract.functions.maxNumServices.return_value.call.return_value = 500

    val1 = c1.max_num_services
    assert val1 == 500
    assert mock_contract.functions.maxNumServices.return_value.call.call_count == 1

    # access on c2
    val2 = c2.max_num_services
    assert val2 == 500
    # Call count should NOT increase
    assert mock_contract.functions.maxNumServices.return_value.call.call_count == 1


def test_epoch_aware_caching(mock_chain_interface):
    """Verify ts_checkpoint caching logic."""
    mock_interface, mock_contract = mock_chain_interface
    ContractCache().clear()
    contract = StakingContract(address="0xEpoch", chain_name="gnosis")

    # Mock return values for tsCheckpoint
    # We use side_effect on the call() method to simulate changing return values if needed
    # But here we just want it to return 1000 once
    mock_contract.functions.tsCheckpoint.return_value.call.return_value = 1000

    # Set liveness period in cache to avoid RPC call for it
    contract._contract_params_cache["livenessPeriod"] = 600

    # 1. Fetch ts_checkpoint
    ts = contract.ts_checkpoint()
    assert ts == 1000
    assert mock_contract.functions.tsCheckpoint.return_value.call.call_count == 1

    # 2. Call again - should be cached
    ts2 = contract.ts_checkpoint()
    assert ts2 == 1000
    assert mock_contract.functions.tsCheckpoint.return_value.call.call_count == 1



"""RPC Monitor for tracking API usage."""

import threading
from collections import defaultdict
from typing import Dict

from iwa.core.utils import configure_logger

logger = configure_logger()


class RPCMonitor:
    """Singleton monitor for tracking RPC usage."""

    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        """Create singleton instance."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(RPCMonitor, cls).__new__(cls)
                    cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        """Initialize monitor."""
        if self._initialized:
            return
        self._counts: Dict[str, int] = defaultdict(int)
        self._lock = threading.Lock()
        self._initialized = True

    def increment(self, metric_name: str, count: int = 1):
        """Increment a metric counter."""
        with self._lock:
            self._counts[metric_name] += count

    def get_counts(self) -> Dict[str, int]:
        """Get a copy of current counts."""
        with self._lock:
            return dict(self._counts)

    def log_stats(self):
        """Log current statistics."""
        stats = self.get_counts()
        if not stats:
            return

        logger.info("RPC Stats Summary:")
        total = 0
        for k, v in sorted(stats.items()):
            logger.info(f"  {k}: {v}")
            total += v
        logger.info(f"  TOTAL: {total}")

    def clear(self):
        """Clear all counters."""
        with self._lock:
            self._counts.clear()



from unittest.mock import MagicMock

from iwa.plugins.olas.service_manager import ServiceManager


def test_service_manager_structure():
    """Verify that ServiceManager has all expected methods from mixins."""
    wallet_mock = MagicMock()
    sm = ServiceManager(wallet=wallet_mock)

    # Check Lifecycle methods
    assert hasattr(sm, "create")
    assert hasattr(sm, "deploy")
    assert hasattr(sm, "spin_up")
    assert hasattr(sm, "wind_down")

    # Check Staking methods
    assert hasattr(sm, "stake")
    assert hasattr(sm, "unstake")
    assert hasattr(sm, "get_staking_status")

    # Check Drain methods
    assert hasattr(sm, "drain_service")
    assert hasattr(sm, "claim_rewards")

    # Check Mech methods
    assert hasattr(sm, "send_mech_request")

    # Check Base methods
    assert hasattr(sm, "get")
    assert hasattr(sm, "_init_contracts")



"""Tests for AccountService."""

from unittest.mock import MagicMock

import pytest

from iwa.core.constants import NATIVE_CURRENCY_ADDRESS
from iwa.core.models import EthereumAddress, StoredSafeAccount
from iwa.core.services.account import AccountService


@pytest.fixture
def mock_key_storage():
    mock = MagicMock()
    mock.master_account = StoredSafeAccount(
        tag="master",
        address="0x1111111111111111111111111111111111111111",
        chains=["gnosis"],
        threshold=1,
        signers=["0x2222222222222222222222222222222222222222"],
    )
    mock.get_account.return_value = MagicMock(address="0x1234")
    mock.get_tag_by_address.return_value = "my_tag"
    mock.accounts = {}
    return mock


@pytest.fixture
def account_service(mock_key_storage):
    return AccountService(mock_key_storage)


def test_init(account_service, mock_key_storage):
    """Test AccountService initialization."""
    assert account_service.key_storage == mock_key_storage


def test_master_account(account_service, mock_key_storage):
    """Test master_account property."""
    result = account_service.master_account

    assert result == mock_key_storage.master_account
    assert result.tag == "master"


def test_get_token_address_native(account_service):
    """Test get_token_address returns native address."""
    mock_chain = MagicMock()

    result = account_service.get_token_address("native", mock_chain)

    assert result == EthereumAddress(NATIVE_CURRENCY_ADDRESS)


def test_get_token_address_valid_address(account_service):
    """Test get_token_address with valid Ethereum address."""
    mock_chain = MagicMock()
    valid_address = "0x1234567890123456789012345678901234567890"

    result = account_service.get_token_address(valid_address, mock_chain)

    assert result == EthereumAddress(valid_address)


def test_get_token_address_by_name(account_service):
    """Test get_token_address resolves token name."""
    mock_chain = MagicMock()
    mock_chain.get_token_address.return_value = EthereumAddress(
        "0x6B175474E89094C44Da98b954EedeAC495271E01"
    )

    result = account_service.get_token_address("DAI", mock_chain)

    assert result == EthereumAddress("0x6B175474E89094C44Da98b954EedeAC495271E01")
    mock_chain.get_token_address.assert_called_with("DAI")


def test_get_token_address_not_found(account_service):
    """Test get_token_address returns None for unknown token."""
    mock_chain = MagicMock()
    mock_chain.name = "gnosis"
    mock_chain.get_token_address.return_value = None

    result = account_service.get_token_address("UNKNOWN", mock_chain)

    assert result is None


def test_resolve_account(account_service, mock_key_storage):
    """Test resolve_account delegates to key_storage."""
    result = account_service.resolve_account("my_tag")

    mock_key_storage.get_account.assert_called_with("my_tag")
    assert result.address == "0x1234"


def test_resolve_account_not_found(account_service, mock_key_storage):
    """Test resolve_account returns None for unknown account."""
    mock_key_storage.get_account.return_value = None

    result = account_service.resolve_account("unknown")

    assert result is None


def test_get_tag_by_address(account_service, mock_key_storage):
    """Test get_tag_by_address delegates to key_storage."""
    result = account_service.get_tag_by_address("0x1234")

    mock_key_storage.get_tag_by_address.assert_called_with("0x1234")
    assert result == "my_tag"


def test_get_account_data(account_service, mock_key_storage):
    """Test get_account_data returns accounts dict."""
    mock_key_storage.accounts = {"0x1234": MagicMock()}

    result = account_service.get_account_data()

    assert result == mock_key_storage.accounts



#!/usr/bin/env python3
"""Simple lifecycle test: create → spin_up → wind_down (no staking)."""

import sys
from pathlib import Path

# Add src to path (scripts are in src/iwa/plugins/olas/scripts/)
sys.path.insert(0, str(Path(__file__).parent.parent.parent.parent.parent))

from iwa.core.wallet import Wallet
from iwa.plugins.olas.service_manager import ServiceManager


def main():
    """Run simple lifecycle integration test: create -> spin_up."""
    print("=" * 60)
    print("  Simple Lifecycle Test: create → spin_up → wind_down")
    print("=" * 60)

    # Initialize wallet
    print("\n🔐 Initializing Wallet...")
    wallet = Wallet()
    print(f"  ✅ Master account: {wallet.master_account.address}")

    # Create service
    print("\n1️⃣ Creating Service...")
    manager = ServiceManager(wallet)
    service_id = manager.create(
        chain_name="gnosis",
        service_name="simple_test_service",
    )

    if not service_id:
        print("  ❌ Failed to create service")
        return False

    print(f"  ✅ Service created with ID: {service_id}")

    # Spin up (activate → register → deploy)
    print("\n2️⃣ Spinning up Service...")
    success = manager.spin_up()
    if not success:
        print("  ❌ Failed to spin up service")
        return False

    print("  ✅ Service deployed!")
    print(f"     - Agent: {manager.service.agent_address}")
    print(f"     - Multisig: {manager.service.multisig_address}")

    # Wind down (terminate → unbond)
    # print("\n3️⃣ Winding down Service...")
    # success = manager.wind_down()
    # if not success:
    #     print("  ❌ Failed to wind down service")
    #     return False

    # print("  ✅ Service wound down successfully!")

    print("\n" + "=" * 60)
    print("  🎉 Full lifecycle completed!")
    print("=" * 60)
    return True


if __name__ == "__main__":
    try:
        success = main()
        sys.exit(0 if success else 1)
    except Exception as e:
        print(f"\n❌ Error: {e}")
        import traceback

        traceback.print_exc()
        sys.exit(1)



"""Chain model definitions."""

from typing import Dict, List, Optional

from pydantic import BaseModel

from iwa.core.models import EthereumAddress
from iwa.core.secrets import secrets
from iwa.core.utils import singleton


class SupportedChain(BaseModel):
    """SupportedChain"""

    name: str
    rpcs: List[str]
    chain_id: int
    native_currency: str
    tokens: Dict[str, EthereumAddress] = {}
    contracts: Dict[str, EthereumAddress] = {}

    @property
    def rpc(self) -> str:
        """Get the primary RPC URL."""
        return self.rpcs[0] if self.rpcs else ""

    def get_token_address(self, token_address_or_name: str) -> Optional[EthereumAddress]:
        """Get token address"""
        if not token_address_or_name:
            return None

        try:
            address = EthereumAddress(token_address_or_name)
        except Exception:
            address = None

        if address and address in self.tokens.values():
            return address

        if address is None:
            # Try direct lookup
            token_addr = self.tokens.get(token_address_or_name, None)
            if token_addr:
                return token_addr

            # Try case-insensitive lookup
            target_lower = token_address_or_name.lower()
            for name, addr in self.tokens.items():
                if name.lower() == target_lower:
                    return addr

            return None

    def get_token_name(self, token_address: str) -> Optional[str]:
        """Get token name from address."""
        addr_lower = token_address.lower()
        for name, addr in self.tokens.items():
            if addr.lower() == addr_lower:
                return name
        return None


@singleton
class Gnosis(SupportedChain):
    """Gnosis Chain"""

    name: str = "Gnosis"
    rpcs: List[str] = []  # Set dynamically in __init__
    chain_id: int = 100
    native_currency: str = "xDAI"
    tokens: Dict[str, EthereumAddress] = {
        "OLAS": EthereumAddress("0xcE11e14225575945b8E6Dc0D4F2dD4C570f79d9f"),
        "WXDAI": EthereumAddress("0xe91D153E0b41518A2Ce8Dd3D7944Fa863463a97d"),
        "USDC": EthereumAddress("0x2a22f9c3b484c3629090FeED35F17Ff8F88f76F0"),
        "SDAI": EthereumAddress("0xaf204776c7245bF4147c2612BF6e5972Ee483701"),
        "EURE": EthereumAddress("0xcB444e90D8198415266c6a2724b7900fb12FC56E"),
    }
    contracts: Dict[str, EthereumAddress] = {
        "GNOSIS_SAFE_MULTISIG_IMPLEMENTATION": EthereumAddress(
            "0x3C1fF68f5aa342D296d4DEe4Bb1cACCA912D95fE"
        ),
        "GNOSIS_SAFE_FALLBACK_HANDLER": EthereumAddress(
            "0xf48f2b2d2a534e402487b3ee7c18c33aec0fe5e4"
        ),
    }

    def __init__(self, **data):
        """Initialize with RPCs from settings (after testing override is applied)."""
        super().__init__(**data)
        if not self.rpcs and secrets.gnosis_rpc:
            self.rpcs = secrets.gnosis_rpc.get_secret_value().split(",")

        # Defensive: ensure no comma-separated strings and NO quotes in list
        new_rpcs = []
        for rpc in self.rpcs:
            parts = [r.strip().strip("'\"") for r in rpc.split(",") if r.strip()]
            new_rpcs.extend(parts)
        self.rpcs = new_rpcs


@singleton
class Ethereum(SupportedChain):
    """Ethereum Mainnet"""

    name: str = "Ethereum"
    rpcs: List[str] = []  # Set dynamically in __init__
    chain_id: int = 1
    native_currency: str = "ETH"
    tokens: Dict[str, EthereumAddress] = {
        "OLAS": EthereumAddress("0x0001A500A6B18995B03f44bb040A5fFc28E45CB0"),
    }
    contracts: Dict[str, EthereumAddress] = {}

    def __init__(self, **data):
        """Initialize with RPCs from settings (after testing override is applied)."""
        super().__init__(**data)
        if not self.rpcs and secrets.ethereum_rpc:
            self.rpcs = secrets.ethereum_rpc.get_secret_value().split(",")

        # Defensive: ensure no comma-separated strings and NO quotes in list
        new_rpcs = []
        for rpc in self.rpcs:
            parts = [r.strip().strip("'\"") for r in rpc.split(",") if r.strip()]
            new_rpcs.extend(parts)
        self.rpcs = new_rpcs


@singleton
class Base(SupportedChain):
    """Base"""

    name: str = "Base"
    rpcs: List[str] = []  # Set dynamically in __init__
    chain_id: int = 8453
    native_currency: str = "ETH"
    tokens: Dict[str, EthereumAddress] = {
        "OLAS": EthereumAddress("0x54330d28ca3357F294334BDC454a032e7f353416"),
    }
    contracts: Dict[str, EthereumAddress] = {}

    def __init__(self, **data):
        """Initialize with RPCs from settings (after testing override is applied)."""
        super().__init__(**data)
        if not self.rpcs and secrets.base_rpc:
            self.rpcs = secrets.base_rpc.get_secret_value().split(",")

        # Defensive: ensure no comma-separated strings and NO quotes in list
        new_rpcs = []
        for rpc in self.rpcs:
            parts = [r.strip().strip("'\"") for r in rpc.split(",") if r.strip()]
            new_rpcs.extend(parts)
        self.rpcs = new_rpcs


@singleton
class SupportedChains:
    """SupportedChains"""

    gnosis: SupportedChain = Gnosis()
    ethereum: SupportedChain = Ethereum()
    base: SupportedChain = Base()



"""Tests for Swap Web API endpoints."""

from unittest.mock import patch

import pytest
from fastapi.testclient import TestClient

# We need to mock Wallet and ChainInterfaces BEFORE importing app from server
# Also mock _get_webui_password to bypass authentication in tests
with (
    patch("iwa.core.wallet.Wallet"),
    patch("iwa.core.chain.ChainInterfaces"),
    patch("iwa.core.wallet.init_db"),
    patch("iwa.web.dependencies._get_webui_password", return_value=None),
):
    from iwa.web.dependencies import verify_auth
    from iwa.web.server import app


# Override auth for all tests
async def override_verify_auth():
    """Override auth for testing."""
    return True


app.dependency_overrides[verify_auth] = override_verify_auth


@pytest.fixture
def client():
    """TestClient for FastAPI app."""
    return TestClient(app)


def test_get_swap_quote_invalid_account(client):
    """Test /api/swap/quote with invalid account."""
    with patch("iwa.web.routers.swap.wallet") as mock_wallet:
        mock_wallet.account_service.resolve_account.return_value = None

        response = client.get(
            "/api/swap/quote?account=invalid&sell_token=WXDAI&buy_token=OLAS&amount=1.0"
        )
        assert response.status_code == 400


def test_swap_tokens_invalid_order_type(client):
    """Test /api/swap with invalid order type."""
    response = client.post(
        "/api/swap",
        json={
            "account": "master",
            "sell_token": "WXDAI",
            "buy_token": "OLAS",
            "amount": 1.0,
            "order_type": "invalid",
            "chain": "gnosis",
        },
    )
    # Should return 422 for validation error
    assert response.status_code == 422


def test_swap_quote_invalid_mode(client):
    """Test /api/swap/quote with invalid mode."""
    response = client.get(
        "/api/swap/quote?account=master&sell_token=WXDAI&buy_token=OLAS&amount=1.0&mode=invalid"
    )
    assert response.status_code == 400


def test_swap_tokens_negative_amount(client):
    """Test /api/swap with negative amount."""
    response = client.post(
        "/api/swap",
        json={
            "account": "master",
            "sell_token": "WXDAI",
            "buy_token": "OLAS",
            "amount": -1.0,
            "order_type": "sell",
            "chain": "gnosis",
        },
    )
    # Should return 422 for validation error, but sometimes 400 depending on handler
    # Accepted behavior is rejection.
    assert response.status_code in (400, 422)


def test_swap_tokens_invalid_chain(client):
    """Test /api/swap with invalid chain."""
    response = client.post(
        "/api/swap",
        json={
            "account": "master",
            "sell_token": "WXDAI",
            "buy_token": "OLAS",
            "amount": 1.0,
            "order_type": "sell",
            "chain": "invalid!chain",
        },
    )
    # Should return 422 for validation error
    assert response.status_code == 422



"""Shared dependencies for Web API routers."""

import secrets
from typing import Optional

from fastapi import Header, HTTPException, Security
from fastapi.security import APIKeyHeader
from loguru import logger

from iwa.core.wallet import Wallet

# Singleton wallet instance (lazy-initialized or injected)
_wallet: Optional[Wallet] = None


def get_wallet() -> Wallet:
    """Get the wallet instance (lazy initialization or injected).

    Returns the injected wallet if set_wallet() was called,
    otherwise creates a new Wallet on first access.
    """
    global _wallet
    if _wallet is None:
        _wallet = Wallet()
    return _wallet


def set_wallet(wallet: Wallet) -> None:
    """Inject an external wallet instance.

    Call this BEFORE importing routers to share a wallet
    with an external application (e.g., Triton).
    """
    global _wallet
    _wallet = wallet


# Backwards compatibility: module-level wallet property
# Deprecated: use get_wallet() instead
class _WalletProxy:
    """Proxy that redirects to get_wallet() for backwards compatibility."""

    def __getattr__(self, name: str):
        return getattr(get_wallet(), name)


wallet = _WalletProxy()

# Authentication
api_key_header = APIKeyHeader(name="X-API-Key", auto_error=False)


def _get_webui_password() -> Optional[str]:
    """Get WEBUI_PASSWORD from secrets (lazy load to ensure secrets.env is loaded)."""
    from iwa.core.secrets import secrets

    if hasattr(secrets, "webui_password") and secrets.webui_password:
        return secrets.webui_password.get_secret_value()
    return None


async def verify_auth(
    x_api_key: Optional[str] = Security(api_key_header), authorization: Optional[str] = Header(None)
) -> bool:
    """Verify authentication via API key or Password.

    Uses timing-safe comparison to prevent timing attacks.
    """
    password = _get_webui_password()

    # If no password configured, allow everything
    if not password:
        return True

    # Check X-API-Key header (timing-safe comparison)
    if x_api_key and secrets.compare_digest(x_api_key, password):
        return True

    # Check Authorization header (Bearer token)
    if authorization:
        scheme, _, param = authorization.partition(" ")
        if scheme.lower() == "bearer" and param and secrets.compare_digest(param, password):
            return True

    raise HTTPException(
        status_code=401,
        detail="Invalid authentication credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )


def get_config():
    """Dependency to provide the Config object."""
    import yaml

    from iwa.core.constants import CONFIG_PATH
    from iwa.core.models import Config

    if not CONFIG_PATH.exists():
        return Config()

    try:
        with open(CONFIG_PATH, "r") as f:
            data = yaml.safe_load(f) or {}
        return Config(**data)
    except Exception as e:
        logger.error(f"Error loading config: {e}")
        return Config()



import sys
from unittest.mock import MagicMock, patch

import pytest

# Mock safe_eth before importing SafeMultisig
sys.modules["safe_eth"] = MagicMock()
sys.modules["safe_eth.eth"] = MagicMock()
sys.modules["safe_eth.eth.constants"] = MagicMock()
sys.modules["safe_eth.safe"] = MagicMock()

from iwa.core.models import StoredSafeAccount
from iwa.plugins.gnosis.safe import SafeMultisig


@pytest.fixture
def mock_secrets():
    with patch("iwa.plugins.gnosis.safe.settings") as mock:
        mock.gnosis_rpc.get_secret_value.return_value = "http://rpc"
        yield mock


@pytest.fixture
def mock_safe_account():
    account = MagicMock(spec=StoredSafeAccount)
    account.address = "0xSafe"
    account.chains = ["gnosis"]
    return account


@pytest.fixture
def mock_safe_lib():
    with (
        patch("iwa.plugins.gnosis.safe.Safe") as mock_safe,
        patch("iwa.plugins.gnosis.safe.EthereumClient") as mock_client,
    ):
        yield mock_safe, mock_client


def test_init_success(mock_safe_account, mock_secrets, mock_safe_lib):
    safe = SafeMultisig(mock_safe_account, "gnosis")
    assert safe.multisig is not None
    mock_safe_lib[1].assert_called_with("http://rpc")  # EthereumClient
    mock_safe_lib[0].assert_called_with("0xSafe", mock_safe_lib[1].return_value)  # Safe


def test_init_invalid_chain(mock_safe_account, mock_secrets):
    with pytest.raises(ValueError, match="Safe account is not deployed on chain"):
        SafeMultisig(mock_safe_account, "ethereum")


def test_get_owners(mock_safe_account, mock_secrets, mock_safe_lib):
    safe = SafeMultisig(mock_safe_account, "gnosis")
    safe.multisig.retrieve_owners.return_value = ["0xOwner"]
    assert safe.get_owners() == ["0xOwner"]


def test_get_threshold(mock_safe_account, mock_secrets, mock_safe_lib):
    safe = SafeMultisig(mock_safe_account, "gnosis")
    safe.multisig.retrieve_threshold.return_value = 2
    assert safe.get_threshold() == 2


def test_get_nonce(mock_safe_account, mock_secrets, mock_safe_lib):
    safe = SafeMultisig(mock_safe_account, "gnosis")
    safe.multisig.retrieve_nonce.return_value = 5
    assert safe.get_nonce() == 5


def test_retrieve_all_info(mock_safe_account, mock_secrets, mock_safe_lib):
    safe = SafeMultisig(mock_safe_account, "gnosis")
    safe.multisig.retrieve_all_info.return_value = {"owners": []}
    assert safe.retrieve_all_info() == {"owners": []}


def test_send_tx_with_callback(mock_safe_account, mock_secrets, mock_safe_lib):
    """Test the new send_tx method with callback."""
    safe = SafeMultisig(mock_safe_account, "gnosis")
    mock_tx = MagicMock()
    safe.multisig.build_multisig_tx.return_value = mock_tx

    def sign_callback(safe_tx):
        return "0xCallbackHash"

    result = safe.send_tx(
        to="0xTo",
        value=100,
        sign_and_execute_callback=sign_callback,
        data="0x1234",
    )

    assert result == "0xCallbackHash"
    safe.multisig.build_multisig_tx.assert_called()
